# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
from unittest import mock

from oslo_utils import uuidutils

from ironic.common import context as ironic_context
from ironic.common import exception
from ironic.common import states
from ironic.common import swift
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules import inspect_utils as utils
from ironic.drivers.modules.inspector.hooks import base as hooks_base
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


@mock.patch('time.sleep', lambda sec: None)
class InspectFunctionTestCase(db_base.DbTestCase):

    def setUp(self):
        super(InspectFunctionTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               boot_interface='pxe')

    @mock.patch.object(utils.LOG, 'info', spec_set=True, autospec=True)
    @mock.patch.object(objects, 'Port', spec_set=True, autospec=True)
    def test_create_ports_if_not_exist(self, port_mock, log_mock):
        macs = {'aa:aa:aa:aa:aa:aa', 'bb:bb:bb:bb:bb:bb'}
        node_id = self.node.id
        port_dict1 = {'address': 'aa:aa:aa:aa:aa:aa', 'node_id': node_id}
        port_dict2 = {'address': 'bb:bb:bb:bb:bb:bb', 'node_id': node_id}
        port_obj1, port_obj2 = mock.MagicMock(), mock.MagicMock()
        port_mock.side_effect = [port_obj1, port_obj2]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            utils.create_ports_if_not_exist(task, macs)
            self.assertTrue(log_mock.called)
            expected_calls = [mock.call(task.context, **port_dict1),
                              mock.call(task.context, **port_dict2)]
            port_mock.assert_has_calls(expected_calls, any_order=True)
            port_obj1.create.assert_called_once_with()
            port_obj2.create.assert_called_once_with()

    @mock.patch.object(utils.LOG, 'warning', spec_set=True, autospec=True)
    @mock.patch.object(utils.LOG, 'info', spec_set=True, autospec=True)
    @mock.patch.object(objects.Port, 'create', spec_set=True, autospec=True)
    def test_create_ports_if_not_exist_mac_exception(self,
                                                     create_mock,
                                                     log_mock,
                                                     warn_mock):
        create_mock.side_effect = exception.MACAlreadyExists('f')
        macs = {'aa:aa:aa:aa:aa:aa', 'bb:bb:bb:bb:bb:bb',
                'aa:aa:aa:aa:aa:aa:bb:bb'}  # WWN
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            utils.create_ports_if_not_exist(task, macs)
        self.assertEqual(2, log_mock.call_count)
        self.assertEqual(2, create_mock.call_count)
        self.assertEqual(1, warn_mock.call_count)

    @mock.patch.object(utils.LOG, 'info', spec_set=True, autospec=True)
    @mock.patch.object(objects, 'Port', spec_set=True, autospec=True)
    def test_create_ports_if_not_exist_attempts_port_creation_blindly(
            self, port_mock, log_info_mock):
        macs = {'aa:bb:cc:dd:ee:ff', 'aa:bb:aa:aa:aa:aa'}
        node_id = self.node.id
        port_dict1 = {'address': 'aa:bb:cc:dd:ee:ff', 'node_id': node_id}
        port_dict2 = {'address': 'aa:bb:aa:aa:aa:aa', 'node_id': node_id}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            utils.create_ports_if_not_exist(task, macs)
            self.assertTrue(log_info_mock.called)
            expected_calls = [mock.call(task.context, **port_dict1),
                              mock.call(task.context, **port_dict2)]
            port_mock.assert_has_calls(expected_calls, any_order=True)
            self.assertEqual(2, port_mock.return_value.create.call_count)


class SwiftCleanUp(db_base.DbTestCase):

    def setUp(self):
        super(SwiftCleanUp, self).setUp()
        self.node = obj_utils.create_test_node(self.context)

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test_clean_up_swift_entries(self, swift_api_mock):
        CONF.set_override('data_backend', 'swift', group='inventory')
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        swift_obj_mock = swift_api_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            utils.clean_up_swift_entries(task)
            object_name = 'inspector_data-' + str(self.node.uuid)
            swift_obj_mock.delete_object.assert_has_calls([
                mock.call(object_name + '-inventory', container),
                mock.call(object_name + '-plugin', container)])

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test_clean_up_swift_entries_with_404_exception(self, swift_api_mock):
        CONF.set_override('data_backend', 'swift', group='inventory')
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        swift_obj_mock = swift_api_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            swift_obj_mock.delete_object.side_effect = [
                exception.SwiftObjectNotFoundError("not found"),
                exception.SwiftObjectNotFoundError("not found")]
            utils.clean_up_swift_entries(task)

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test_clean_up_swift_entries_with_fail_exception(self, swift_api_mock):
        CONF.set_override('data_backend', 'swift', group='inventory')
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        swift_obj_mock = swift_api_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            swift_obj_mock.delete_object.side_effect = [
                exception.SwiftOperationError("failed"),
                exception.SwiftObjectNotFoundError("not found")]
            self.assertRaises(exception.SwiftObjectStillExists,
                              utils.clean_up_swift_entries, task)

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test_clean_up_swift_entries_with_fail_exceptions(self, swift_api_mock):
        CONF.set_override('data_backend', 'swift', group='inventory')
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        swift_obj_mock = swift_api_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            swift_obj_mock.delete_object.side_effect = [
                exception.SwiftOperationError("failed"),
                exception.SwiftOperationError("failed")]
            self.assertRaises((exception.SwiftObjectStillExists,
                               exception.SwiftObjectStillExists),
                              utils.clean_up_swift_entries, task)


class IntrospectionDataStorageFunctionsTestCase(db_base.DbTestCase):
    fake_inventory_data = {"cpu": "amd"}
    fake_plugin_data = {"disks": [{"name": "/dev/vda"}]}

    def setUp(self):
        super(IntrospectionDataStorageFunctionsTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context)

    def test_store_inspection_data_db(self):
        CONF.set_override('data_backend', 'database', group='inventory')
        fake_context = ironic_context.RequestContext()
        utils.store_inspection_data(self.node, self.fake_inventory_data,
                                    self.fake_plugin_data, fake_context)
        stored = objects.NodeInventory.get_by_node_id(self.context,
                                                      self.node.id)
        self.assertEqual(self.fake_inventory_data, stored["inventory_data"])
        self.assertEqual(self.fake_plugin_data, stored["plugin_data"])

    @mock.patch.object(utils, '_store_inspection_data_in_swift',
                       autospec=True)
    def test_store_inspection_data_swift(self, mock_store_data):
        CONF.set_override('data_backend', 'swift', group='inventory')
        CONF.set_override(
            'swift_data_container', 'inspection_data',
            group='inventory')
        fake_context = ironic_context.RequestContext()
        utils.store_inspection_data(self.node, self.fake_inventory_data,
                                    self.fake_plugin_data, fake_context)
        mock_store_data.assert_called_once_with(
            self.node.uuid, inventory_data=self.fake_inventory_data,
            plugin_data=self.fake_plugin_data)

    def test_store_inspection_data_nostore(self):
        CONF.set_override('data_backend', 'none', group='inventory')
        fake_context = ironic_context.RequestContext()
        utils.store_inspection_data(self.node, self.fake_inventory_data,
                                    self.fake_plugin_data, fake_context)
        self.assertRaises(exception.NodeInventoryNotFound,
                          objects.NodeInventory.get_by_node_id,
                          self.context, self.node.id)

    def test_get_inspection_data_db(self):
        CONF.set_override('data_backend', 'database', group='inventory')
        obj_utils.create_test_inventory(
            self.context, self.node,
            inventory_data=self.fake_inventory_data,
            plugin_data=self.fake_plugin_data)
        fake_context = ironic_context.RequestContext()
        ret = utils.get_inspection_data(self.node, fake_context)
        fake_inspection_data = {'inventory': self.fake_inventory_data,
                                'plugin_data': self.fake_plugin_data}
        self.assertEqual(ret, fake_inspection_data)

    def test_get_inspection_data_db_exception(self):
        CONF.set_override('data_backend', 'database', group='inventory')
        fake_context = ironic_context.RequestContext()
        self.assertRaises(
            exception.NodeInventoryNotFound, utils.get_inspection_data,
            self.node, fake_context)

    @mock.patch.object(utils, '_get_inspection_data_from_swift', autospec=True)
    def test_get_inspection_data_swift(self, mock_get_data):
        CONF.set_override('data_backend', 'swift', group='inventory')
        CONF.set_override(
            'swift_data_container', 'inspection_data',
            group='inventory')
        fake_context = ironic_context.RequestContext()
        ret = utils.get_inspection_data(self.node, fake_context)
        mock_get_data.assert_called_once_with(self.node.uuid)
        self.assertEqual(mock_get_data.return_value, ret)

    @mock.patch.object(utils, '_get_inspection_data_from_swift', autospec=True)
    def test_get_inspection_data_swift_exception(self, mock_get_data):
        CONF.set_override('data_backend', 'swift', group='inventory')
        CONF.set_override(
            'swift_data_container', 'inspection_data',
            group='inventory')
        fake_context = ironic_context.RequestContext()
        mock_get_data.side_effect = exception.SwiftObjectNotFoundError()
        self.assertRaises(
            exception.NodeInventoryNotFound, utils.get_inspection_data,
            self.node, fake_context)

    def test_get_inspection_data_nostore(self):
        CONF.set_override('data_backend', 'none', group='inventory')
        fake_context = ironic_context.RequestContext()
        self.assertRaises(
            exception.NodeInventoryNotFound, utils.get_inspection_data,
            self.node, fake_context)

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test__store_inspection_data_in_swift(self, swift_api_mock):
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        utils._store_inspection_data_in_swift(
            self.node.uuid, self.fake_inventory_data, self.fake_plugin_data)
        swift_obj_mock = swift_api_mock.return_value
        object_name = 'inspector_data-' + str(self.node.uuid)
        swift_obj_mock.create_object_from_data.assert_has_calls([
            mock.call(object_name + '-inventory', self.fake_inventory_data,
                      container),
            mock.call(object_name + '-plugin', self.fake_plugin_data,
                      container)])

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test__get_inspection_data_from_swift(self, swift_api_mock):
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        swift_obj_mock = swift_api_mock.return_value
        swift_obj_mock.get_object.side_effect = [
            self.fake_inventory_data,
            self.fake_plugin_data
        ]
        ret = utils._get_inspection_data_from_swift(self.node.uuid)
        req_ret = {"inventory": self.fake_inventory_data,
                   "plugin_data": self.fake_plugin_data}
        self.assertEqual(req_ret, ret)

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test__get_inspection_data_from_swift_exception(self, swift_api_mock):
        container = 'inspection_data'
        CONF.set_override('swift_data_container', container, group='inventory')
        swift_obj_mock = swift_api_mock.return_value
        swift_obj_mock.get_object.side_effect = [
            exception.SwiftOperationError,
            self.fake_plugin_data
        ]
        self.assertRaises(exception.SwiftObjectNotFoundError,
                          utils._get_inspection_data_from_swift,
                          self.node.uuid)


class LookupNodeTestCase(db_base.DbTestCase):

    def setUp(self):
        super().setUp()
        self.bmc = '192.0.2.1'
        self.node = obj_utils.create_test_node(
            self.context,
            driver_internal_info={utils.LOOKUP_CACHE_FIELD: [self.bmc]},
            provision_state=states.INSPECTWAIT)

        self.macs = ['11:22:33:44:55:66', '12:34:56:78:90:ab']
        self.unknown_mac = '66:55:44:33:22:11'
        self.ports = [
            obj_utils.create_test_port(self.context,
                                       uuid=uuidutils.generate_uuid(),
                                       node_id=self.node.id,
                                       address=addr)
            for addr in self.macs
        ]

        self.bmc2 = '1.2.1.2'
        self.mac2 = '00:11:00:11:00:11'
        self.node2 = obj_utils.create_test_node(
            self.context,
            uuid=uuidutils.generate_uuid(),
            driver_internal_info={utils.LOOKUP_CACHE_FIELD: [self.bmc2]},
            provision_state=states.INSPECTWAIT)
        obj_utils.create_test_port(self.context,
                                   node_id=self.node2.id,
                                   address=self.mac2)

    def test_no_input(self):
        self.assertRaises(exception.BadRequest, utils.lookup_node,
                          self.context, [], [], None)

    def test_by_macs(self):
        result = utils.lookup_node(self.context, self.macs[::-1], [], None)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_macs_partial(self):
        macs = [self.macs[1], self.unknown_mac]
        result = utils.lookup_node(self.context, macs, [], None)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_mac_not_found(self):
        self.assertRaises(utils.AutoEnrollPossible, utils.lookup_node,
                          self.context, [self.unknown_mac], [], None)

    def test_by_mac_wrong_state(self):
        self.node.provision_state = states.AVAILABLE
        self.node.save()
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, self.macs, [], None)

    def test_conflicting_macs(self):
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, [self.macs[0], self.mac2], [], None)

    def test_by_bmc(self):
        result = utils.lookup_node(self.context, [], ['192.0.2.1'], None)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_bmc_and_mac(self):
        result = utils.lookup_node(
            self.context, [self.macs[0]], [self.bmc], None)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_unknown_bmc_and_mac(self):
        result = utils.lookup_node(
            self.context, [self.unknown_mac], [self.bmc], None)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_bmc_and_mac_and_uuid(self):
        result = utils.lookup_node(
            self.context, [self.macs[0]], [self.bmc], self.node.uuid)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_bmc_not_found(self):
        self.assertRaises(utils.AutoEnrollPossible, utils.lookup_node,
                          self.context, [], ['192.168.1.1'], None)

    def test_by_bmc_and_mac_not_found(self):
        self.assertRaises(utils.AutoEnrollPossible, utils.lookup_node,
                          self.context, [self.unknown_mac],
                          ['192.168.1.1'], None)

    def test_by_bmc_wrong_state(self):
        self.node.provision_state = states.AVAILABLE
        self.node.save()
        # Limitation of auto-discovery: cannot de-duplicate nodes by BMC
        # addresses only. Should not happen too often in reality.
        # If it does happen, auto-discovery will create a duplicate node.
        self.assertRaises(utils.AutoEnrollPossible, utils.lookup_node,
                          self.context, [], [self.bmc], None)

    def test_conflicting_macs_and_bmc(self):
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, self.macs, [self.bmc2], None)

    def test_duplicate_bmc(self):
        # This can happen with Redfish. There is no way to resolve the conflict
        # other than by using MACs.
        obj_utils.create_test_node(
            self.context,
            uuid=uuidutils.generate_uuid(),
            driver_internal_info={utils.LOOKUP_CACHE_FIELD: [self.bmc]},
            provision_state=states.INSPECTWAIT)
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, [], [self.bmc], None)

    def test_duplicate_bmc_and_unknown_mac(self):
        obj_utils.create_test_node(
            self.context,
            uuid=uuidutils.generate_uuid(),
            driver_internal_info={utils.LOOKUP_CACHE_FIELD: [self.bmc]},
            provision_state=states.INSPECTWAIT)
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, [self.unknown_mac], [self.bmc], None)

    def test_duplicate_bmc_resolved_by_macs(self):
        obj_utils.create_test_node(
            self.context,
            uuid=uuidutils.generate_uuid(),
            driver_internal_info={utils.LOOKUP_CACHE_FIELD: [self.bmc]},
            provision_state=states.INSPECTWAIT)
        result = utils.lookup_node(
            self.context, [self.macs[0]], [self.bmc], None)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_uuid(self):
        result = utils.lookup_node(self.context, [], [], self.node.uuid)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_uuid_and_unknown_macs(self):
        result = utils.lookup_node(
            self.context, [self.unknown_mac], [], self.node.uuid)
        self.assertEqual(self.node.uuid, result.uuid)

    def test_by_uuid_not_found(self):
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, [], [], uuidutils.generate_uuid())

    def test_by_uuid_wrong_state(self):
        self.node.provision_state = states.AVAILABLE
        self.node.save()
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, [], [], self.node.uuid)

    def test_conflicting_macs_and_uuid(self):
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, self.macs, [], self.node2.uuid)

    def test_conflicting_bmc_and_uuid(self):
        self.assertRaises(exception.NotFound, utils.lookup_node,
                          self.context, self.macs, [self.bmc], self.node2.uuid)


class GetBMCAddressesTestCase(db_base.DbTestCase):

    def test_localhost_ignored(self):
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'ipmi_address': '127.0.0.1'})
        self.assertEqual(set(), utils._get_bmc_addresses(node))

    def test_localhost_as_url_ignored(self):
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'redfish_address': 'https://localhost/redfish'})
        self.assertEqual(set(), utils._get_bmc_addresses(node))

    def test_normal_ip(self):
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'ipmi_address': '192.0.2.1'})
        self.assertEqual({'192.0.2.1'}, utils._get_bmc_addresses(node))

    def test_normal_ip_as_url(self):
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'redfish_address': 'https://192.0.2.1/redfish'})
        self.assertEqual({'192.0.2.1'}, utils._get_bmc_addresses(node))

    def test_normal_ipv6_as_url(self):
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'redfish_address': 'https://[2001:db8::42]/redfish'})
        self.assertEqual({'2001:db8::42'}, utils._get_bmc_addresses(node))

    @mock.patch.object(socket, 'getaddrinfo', autospec=True)
    def test_resolved_host(self, mock_getaddrinfo):
        mock_getaddrinfo.return_value = [
            (socket.AF_INET6, socket.SOCK_STREAM, socket.SOL_TCP,
             '', ('2001:db8::42', None)),
            (socket.AF_INET, socket.SOCK_STREAM, socket.SOL_TCP,
             '', ('192.0.2.1', None)),
        ]
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'ipmi_address': 'example.com'})
        self.assertEqual({'example.com', '192.0.2.1', '2001:db8::42'},
                         utils._get_bmc_addresses(node))
        mock_getaddrinfo.assert_called_once_with(
            'example.com', None, proto=socket.SOL_TCP)

    @mock.patch.object(socket, 'getaddrinfo', autospec=True)
    def test_resolved_host_in_url(self, mock_getaddrinfo):
        mock_getaddrinfo.return_value = [
            (socket.AF_INET6, socket.SOCK_STREAM, socket.SOL_TCP,
             '', ('2001:db8::42', None)),
            (socket.AF_INET, socket.SOCK_STREAM, socket.SOL_TCP,
             '', ('192.0.2.1', None)),
        ]
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'redfish_address': 'https://example.com:8080/v1'})
        self.assertEqual({'example.com', '192.0.2.1', '2001:db8::42'},
                         utils._get_bmc_addresses(node))
        mock_getaddrinfo.assert_called_once_with(
            'example.com', None, proto=socket.SOL_TCP)

    def test_redfish_bmc_address_ipv6_brackets_no_scheme(self):
        node = obj_utils.create_test_node(
            self.context,
            driver_info={'redfish_address': '[2001:db8::42]'})
        self.assertEqual({'2001:db8::42'}, utils._get_bmc_addresses(node))


class LookupCacheTestCase(db_base.DbTestCase):

    def setUp(self):
        super().setUp()
        self.bmc = '192.0.2.1'
        self.node = obj_utils.create_test_node(
            self.context,
            driver_internal_info={utils.LOOKUP_CACHE_FIELD: [self.bmc]},
            provision_state=states.INSPECTWAIT)

    def test_clear(self):
        result = utils.clear_lookup_addresses(self.node)
        self.assertEqual([self.bmc], result)
        self.assertEqual({}, self.node.driver_internal_info)

    @mock.patch.object(utils, '_get_bmc_addresses', autospec=True)
    def test_new_value(self, mock_get_addr):
        mock_get_addr.return_value = {'192.0.2.42'}
        utils.cache_lookup_addresses(self.node)
        self.assertEqual({utils.LOOKUP_CACHE_FIELD: ['192.0.2.42']},
                         self.node.driver_internal_info)

    @mock.patch.object(utils, '_get_bmc_addresses', autospec=True)
    def test_replace_with_empty(self, mock_get_addr):
        mock_get_addr.return_value = set()
        utils.cache_lookup_addresses(self.node)
        self.assertEqual({}, self.node.driver_internal_info)


class RunInspectionHooksTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(
            self.context,
            inspect_interface='agent',
            provision_state=states.INSPECTING)
        self.task = mock.Mock(spec=task_manager.TaskManager, node=self.node)
        self.hooks = [
            mock.Mock(name=str(i),
                      obj=mock.MagicMock(spec=hooks_base.InspectionHook))
            for i in range(2)
        ]
        self.on_error_plugin_data = mock.MagicMock()
        self.inventory = {"interfaces": [42]}
        self.plugin_data = {"logs": "abcd"}

    def test_no_on_error(self):
        utils.run_inspection_hooks(self.task, self.inventory,
                                   self.plugin_data, self.hooks,
                                   None)
        for hook in self.hooks:
            hook.obj.preprocess.assert_called_once_with(
                self.task, self.inventory, self.plugin_data)
            hook.obj.assert_called_once_with(
                self.task, self.inventory, self.plugin_data)

        self.on_error_plugin_data.assert_not_called()

    def test_pre_hook_on_error_callback(self):
        failing_pre_hook = mock.MagicMock(spec=hooks_base.InspectionHook)
        failing_pre_hook.preprocess.side_effect = (
            exception.HardwareInspectionFailure("fake exception")
        )
        failing_hooks = self.hooks.copy()
        failing_hooks.append(mock.Mock(name="failing", obj=failing_pre_hook))

        self.assertRaises(exception.HardwareInspectionFailure,
                          utils.run_inspection_hooks,
                          self.task, self.inventory, self.plugin_data,
                          failing_hooks,
                          self.on_error_plugin_data)

        self.on_error_plugin_data.assert_called_once()

    def test_post_hook_on_error_callback(self):
        failing_post_hook = mock.MagicMock(spec=hooks_base.InspectionHook)
        failing_post_hook.__call__ = mock.Mock(
            side_effect=exception.HardwareInspectionFailure("fake exception")
        )
        failing_hooks = self.hooks.copy()
        failing_hooks.append(mock.Mock(name="failing", obj=failing_post_hook))

        self.assertRaises(exception.HardwareInspectionFailure,
                          utils.run_inspection_hooks,
                          self.task, self.inventory, self.plugin_data,
                          failing_hooks,
                          self.on_error_plugin_data)

        self.on_error_plugin_data.assert_called_once()



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_db import sqlalchemy
from oslo_upgradecheck.upgradecheck import Code
from sqlalchemy.engine import url as sa_url

from ironic.cmd import dbsync
from ironic.cmd import status
from ironic.tests.unit.db import base as db_base


class TestUpgradeChecks(db_base.DbTestCase):

    def setUp(self):
        super(TestUpgradeChecks, self).setUp()
        self.cmd = status.Checks()

    def test__check_obj_versions(self):
        check_result = self.cmd._check_obj_versions()
        self.assertEqual(Code.SUCCESS, check_result.code)

    @mock.patch.object(dbsync.DBCommand, 'check_obj_versions', autospec=True)
    def test__check_obj_versions_bad(self, mock_check):
        msg = 'This is bad'
        mock_check.return_value = msg
        check_result = self.cmd._check_obj_versions()
        self.assertEqual(Code.FAILURE, check_result.code)
        self.assertEqual(msg, check_result.details)

    def test__check_allocations_table_ok(self):
        check_result = self.cmd._check_allocations_table()
        self.assertEqual(Code.SUCCESS,
                         check_result.code)

    def _create_mock_context_manager(self, return_self=True):
        mock_obj = mock.Mock()
        if return_self:
            val = mock_obj
        else:
            val = mock.Mock()
        mock_obj.__enter__ = mock.Mock(return_value=val)
        mock_obj.__exit__ = mock.Mock()
        return mock_obj

    @mock.patch.object(sqlalchemy.enginefacade.reader,
                       'get_engine', autospec=True)
    def test__check_allocations_table_latin1(self, mock_reader):
        mock_engine = mock.Mock()
        mock_res = mock.Mock()
        mock_engine.url = sa_url.make_url(
            'mysql+pymysql://ironic:pass@192.0.2.10/ironic')
        mock_res.all.return_value = (
            '... ENGINE=InnoDB DEFAULT CHARSET=latin1',
        )
        mock_conn = self._create_mock_context_manager(True)
        mock_trans = self._create_mock_context_manager(False)
        mock_engine.connect.return_value = mock_conn
        mock_conn.execute.return_value = mock_res
        mock_conn.begin.return_value = mock_trans
        mock_reader.return_value = mock_engine
        check_result = self.cmd._check_allocations_table()
        self.assertEqual(Code.WARNING,
                         check_result.code)
        expected_msg = ('The Allocations table is is not using UTF8 '
                        'encoding. This is corrected in later versions '
                        'of Ironic, where the table character set schema '
                        'is automatically migrated. Continued use of a '
                        'non-UTF8 character set may produce unexpected '
                        'results.')
        self.assertEqual(expected_msg, check_result.details)

    @mock.patch.object(sqlalchemy.enginefacade.reader,
                       'get_engine', autospec=True)
    def test__check_allocations_table_myiasm(self, mock_reader):
        mock_engine = mock.Mock()
        mock_res = mock.Mock()
        mock_engine.url = sa_url.make_url(
            'mysql+pymysql://ironic:pass@192.0.2.10/ironic')
        mock_res.all.return_value = (
            '... ENGINE=MyIASM DEFAULT CHARSET=utf8',
        )
        mock_conn = self._create_mock_context_manager(True)
        mock_trans = self._create_mock_context_manager(False)
        mock_engine.connect.return_value = mock_conn
        mock_conn.execute.return_value = mock_res
        mock_conn.begin.return_value = mock_trans
        mock_reader.return_value = mock_engine
        check_result = self.cmd._check_allocations_table()
        self.assertEqual(Code.WARNING,
                         check_result.code)
        expected_msg = ('The engine used by MySQL for the allocations '
                        'table is not the intended engine for the Ironic '
                        'database tables to use. This may have been a '
                        'result of an error with the table creation schema. '
                        'This may require Database Administrator '
                        'intervention and downtime to dump, modify the '
                        'table engine to utilize InnoDB, and reload the '
                        'allocations table to utilize the InnoDB engine.')
        self.assertEqual(expected_msg, check_result.details)

    @mock.patch.object(sqlalchemy.enginefacade.reader,
                       'get_engine', autospec=True)
    def test__check_allocations_table_myiasm_both(self, mock_reader):
        mock_engine = mock.Mock()
        mock_res = mock.Mock()
        mock_engine.url = sa_url.make_url(
            'mysql+pymysql://ironic:pass@192.0.2.10/ironic')
        mock_res.all.return_value = (
            '... ENGINE=MyIASM DEFAULT CHARSET=latin1',
        )
        mock_conn = self._create_mock_context_manager(True)
        mock_trans = self._create_mock_context_manager(False)
        mock_engine.connect.return_value = mock_conn
        mock_conn.execute.return_value = mock_res
        mock_conn.begin.return_value = mock_trans
        mock_reader.return_value = mock_engine
        check_result = self.cmd._check_allocations_table()
        self.assertEqual(Code.WARNING,
                         check_result.code)
        expected_msg = ('The Allocations table is is not using UTF8 '
                        'encoding. This is corrected in later versions '
                        'of Ironic, where the table character set schema '
                        'is automatically migrated. Continued use of a '
                        'non-UTF8 character set may produce unexpected '
                        'results. Additionally: '
                        'The engine used by MySQL for the allocations '
                        'table is not the intended engine for the Ironic '
                        'database tables to use. This may have been a '
                        'result of an error with the table creation schema. '
                        'This may require Database Administrator '
                        'intervention and downtime to dump, modify the '
                        'table engine to utilize InnoDB, and reload the '
                        'allocations table to utilize the InnoDB engine.')
        self.assertEqual(expected_msg, check_result.details)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
import time

from oslo_log import log
from tooz import hashring

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import utils
from ironic.conf import CONF
from ironic.db import api as dbapi


LOG = log.getLogger(__name__)


class HashRingManager(object):
    _hash_rings = (None, 0)
    _lock = threading.Lock()

    def __init__(self, use_groups=True, cache=True):
        self.dbapi = dbapi.get_instance()
        self.use_groups = use_groups
        self.cache = cache

    @property
    def ring(self):
        interval = CONF.hash_ring_reset_interval
        limit = time.monotonic() - interval

        if not self.cache:
            return self._load_hash_rings()

        # Hot path, no lock. Using a local variable to avoid races with code
        # changing the class variable.
        hash_rings, updated_at = self.__class__._hash_rings
        if (hash_rings is not None
            and (updated_at >= limit
                 or utils.is_ironic_using_sqlite())):
            # Returning the hash ring for us, if it is still valid,
            # or if we're using sqlite.
            return hash_rings

        with self._lock:
            hash_rings, updated_at = self.__class__._hash_rings
            if hash_rings is None or updated_at < limit:
                LOG.debug('Rebuilding cached hash rings')
                hash_rings = self._load_hash_rings()
                self.__class__._hash_rings = hash_rings, time.monotonic()
                LOG.debug('Finished rebuilding hash rings, available drivers '
                          'are %s', ', '.join(hash_rings))
            return hash_rings

    def _load_hash_rings(self):
        rings = {}
        d2c = self.dbapi.get_active_hardware_type_dict(
            use_groups=self.use_groups)

        for driver_name, hosts in d2c.items():
            rings[driver_name] = hashring.HashRing(
                hosts, partitions=2 ** CONF.hash_partition_exponent,
                hash_function=CONF.hash_ring_algorithm)

        return rings

    @classmethod
    def reset(cls):
        with cls._lock:
            LOG.debug('Resetting cached hash rings')
            cls._hash_rings = (None, 0)

    def get_ring(self, driver_name, conductor_group):
        try:
            return self._get_ring(driver_name, conductor_group)
        except (exception.DriverNotFound, exception.TemporaryFailure):
            # NOTE(dtantsur): we assume that this case is more often caused by
            # conductors coming and leaving, so we try to rebuild the rings.
            LOG.debug('No conductor from group %(group)s found for driver '
                      '%(driver)s, trying to rebuild the hash rings',
                      {'driver': driver_name,
                       'group': conductor_group or '<none>'})

        self.__class__.reset()
        return self._get_ring(driver_name, conductor_group)

    def _get_ring(self, driver_name, conductor_group):
        # There are no conductors, temporary failure - 503 Service Unavailable
        ring = self.ring  # a property, don't load twice
        if not ring:
            raise exception.TemporaryFailure()

        try:
            if self.use_groups:
                return ring['%s:%s' % (conductor_group, driver_name)]
            return ring[driver_name]
        except KeyError:
            raise exception.DriverNotFound(
                _("The driver '%s' is unknown.") % driver_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg
from oslo_utils import strutils

from ironic.common import exception
from ironic.common import rpc
from ironic.objects import base
from ironic.objects import fields


CONF = cfg.CONF


# Definition of notification levels in increasing order of severity
NOTIFY_LEVELS = {
    fields.NotificationLevel.DEBUG: 0,
    fields.NotificationLevel.INFO: 1,
    fields.NotificationLevel.WARNING: 2,
    fields.NotificationLevel.ERROR: 3,
    fields.NotificationLevel.CRITICAL: 4
}


@base.IronicObjectRegistry.register
class EventType(base.IronicObject):
    """Defines the event_type to be sent on the wire.

    An EventType must specify the object being acted on, a string describing
    the action being taken on the notification, and the status of the action.
    """
    # Version 1.0: Initial version
    # Version 1.1: "phase" field was renamed to "status" and only accepts
    #              "start", "end", "error", or "success" as valid
    VERSION = '1.1'

    fields = {
        'object': fields.StringField(nullable=False),
        'action': fields.StringField(nullable=False),
        'status': fields.NotificationStatusField()
    }

    def to_event_type_field(self):
        """Constructs string for event_type to be sent on the wire.

           The string is in the format: baremetal.<object>.<action>.<status>

           :raises: ValueError if self.status is not one of
                    :class:`fields.NotificationStatusField`
           :returns: event_type string
        """
        parts = ['baremetal', self.object, self.action, self.status]
        return '.'.join(parts)


# NOTE(mariojv) This class will not be used directly and is just a base class
# for notifications, so we don't need to register it.
@base.IronicObjectRegistry.register_if(False)
class NotificationBase(base.IronicObject):
    """Base class for versioned notifications.

    Subclasses must define the "payload" field, which must be a subclass of
    NotificationPayloadBase.
    """
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'level': fields.NotificationLevelField(),
        'event_type': fields.ObjectField('EventType'),
        'publisher': fields.ObjectField('NotificationPublisher')
    }

    # NOTE(mariojv) This may be a candidate for something oslo.messaging
    # implements instead of in ironic.
    def _should_notify(self):
        """Determine whether the notification should be sent.

        A notification is sent when the level of the notification is
        greater than or equal to the level specified in the
        configuration, in the increasing order of DEBUG, INFO, WARNING,
        ERROR, CRITICAL.

        :return: True if notification should be sent, False otherwise.
        """
        if CONF.notification_level is None:
            return False
        return (NOTIFY_LEVELS[self.level]
                >= NOTIFY_LEVELS[CONF.notification_level])

    def emit(self, context):
        """Send the notification.

           :raises: NotificationPayloadError
           :raises: oslo_versionedobjects.exceptions.MessageDeliveryFailure
        """
        if not self._should_notify():
            return
        if not self.payload.populated:
            raise exception.NotificationPayloadError(
                class_name=self.__class__.__name__)
        # NOTE(mariojv) By default, oslo_versionedobjects includes a list
        # of "changed fields" for the object in the output of
        # obj_to_primitive. This is unneeded since every field of the
        # object will look changed, since each payload is a newly created
        # object, so we drop the changes.
        self.payload.obj_reset_changes()
        event_type = self.event_type.to_event_type_field()
        publisher_id = '%s.%s' % (self.publisher.service, self.publisher.host)
        payload = self.payload.obj_to_primitive()

        notifier = rpc.get_versioned_notifier(publisher_id)
        notify = getattr(notifier, self.level)
        notify(context, event_type=event_type, payload=payload)


# NOTE(mariojv) This class will not be used directly and is just a base class
# for notifications, so we don't need to register it.
@base.IronicObjectRegistry.register_if(False)
class NotificationPayloadBase(base.IronicObject):
    """Base class for the payload of versioned notifications."""

    SCHEMA = {}
    # Version 1.0: Initial version
    VERSION = '1.0'

    def __init__(self, *args, **kwargs):
        super(NotificationPayloadBase, self).__init__(*args, **kwargs)
        # If SCHEMA is empty, the payload is already populated
        self.populated = not self.SCHEMA

    def populate_schema(self, **kwargs):
        """Populate the object based on the SCHEMA and the source objects

        :param kwargs: A dict contains the source object and the keys defined
                       in the SCHEMA
        :raises: NotificationSchemaObjectError
        :raises: NotificationSchemaKeyError
        """
        for key, (obj, field) in self.SCHEMA.items():
            try:
                source = kwargs[obj]
            except KeyError:
                raise exception.NotificationSchemaObjectError(obj=obj,
                                                              source=kwargs)
            try:
                setattr(self, key, getattr(source, field))
            except NotImplementedError:
                # The object is missing (a value for) field. Oslo try to load
                # value via obj_load_attr() method which is not implemented.
                # If this field is nullable in this payload, set its payload
                # value to None.
                field_obj = self.fields.get(key)
                if field_obj is not None and getattr(field_obj, 'nullable',
                                                     False):
                    setattr(self, key, None)
                    continue
                raise exception.NotificationSchemaKeyError(obj=obj,
                                                           field=field,
                                                           key=key)
            except Exception:
                raise exception.NotificationSchemaKeyError(obj=obj,
                                                           field=field,
                                                           key=key)
        self.populated = True


@base.IronicObjectRegistry.register
class NotificationPublisher(base.IronicObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'service': fields.StringField(nullable=False),
        'host': fields.StringField(nullable=False)
    }


def mask_secrets(payload):
    """Remove secrets from payload object."""
    mask = '******'

    dict_fields = ['instance_info', 'driver_info', 'driver_internal_info']
    for f in dict_fields:
        if hasattr(payload, f):
            masked = strutils.mask_dict_password(getattr(payload, f), mask)
            setattr(payload, f, masked)

    if hasattr(payload, 'instance_info'):
        if 'image_url' in payload.instance_info:
            payload.instance_info['image_url'] = mask



#   Copyright 2025 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

# NOTE(TheJulia): This file is based upon, in part, some of the TripleO
# project container uploader.
# https://github.com/openstack-archive/tripleo-common/blame/stable/wallaby/tripleo_common/image/image_uploader.py

import base64
import json
import re
import requests
from requests import auth as requests_auth
import tenacity
from urllib import parse

from oslo_log import log as logging

from ironic.common import checksum_utils
from ironic.common import exception
from ironic.common.i18n import _
from ironic.conf import CONF

LOG = logging.getLogger(__name__)


(
    CALL_MANIFEST,
    CALL_BLOB,
    CALL_TAGS,
) = (
    '%(image)s/manifests/%(tag)s',
    '%(image)s/blobs/%(digest)s',
    '%(image)s/tags/list',
)

(
    MEDIA_OCI_MANIFEST_V1,
    MEDIA_OCI_INDEX_V1,
) = (
    'application/vnd.oci.image.manifest.v1+json',
    'application/vnd.oci.image.index.v1+json',
)


class MakeSession(object):
    """Class method to uniformly create sessions.

    Sessions created by this class will retry on errors with an exponential
    backoff before raising an exception. Because our primary interaction is
    with the container registries the adapter will also retry on 401 and
    404. This is being done because registries commonly return 401 when an
    image is not found, which is commonly a cache miss. See the adapter
    definitions for more on retry details.
    """
    def __init__(self, verify=True):
        self.session = requests.Session()
        self.session.verify = verify

    def create(self):
        return self.__enter__()

    def __enter__(self):
        return self.session

    def __exit__(self, *args, **kwargs):
        self.session.close()


class RegistrySessionHelper(object):
    """Class with various registry session helpers

    This class contains a bunch of static methods to be used when making
    session requests against a container registry. The methods are primarily
    used to handle authentication/reauthentication for the requests against
    registries that require auth.
    """
    @staticmethod
    def check_status(session, request, allow_reauth=True):
        """Check request status and trigger reauth

        This function can be used to check if we need to perform authentication
        for a container registry request because we've gotten a 401.
        """
        text = getattr(request, 'text', 'unknown')
        reason = getattr(request, 'reason', 'unknown')
        status_code = getattr(request, 'status_code', None)
        headers = getattr(request, 'headers', {})

        if status_code >= 300 and status_code != 401:
            LOG.info(
                'OCI client got a Non-2xx: status %s, reason %s, text %s',
                status_code,
                reason,
                text)

        if status_code == 401:
            LOG.warning(
                'OCI client failed: status %s, reason %s text %s',
                status_code,
                reason,
                text)
            www_auth = headers.get(
                'www-authenticate',
                headers.get(
                    'Www-Authenticate'
                )
            )
            if www_auth:
                error = None
                # Handle docker.io shenanigans. docker.io will return 401
                # for 403 and 404 but provide an error string. Other registries
                # like registry.redhat.io and quay.io do not do this. So if
                # we find an error string, check to see if we should reauth.
                do_reauth = allow_reauth
                if 'error=' in www_auth:
                    error = re.search('error="(.*?)"', www_auth).group(1)
                    LOG.warning(
                        'Error detected in auth headers: error %s', error)
                    do_reauth = (error == 'invalid_token' and allow_reauth)
                if do_reauth:
                    if hasattr(session, 'reauthenticate'):
                        # This is a re-auth counter
                        reauth = int(session.headers.get('_ReAuth', 0))
                        reauth += 1
                        session.headers['_ReAuth'] = str(reauth)
                        session.reauthenticate(**session.auth_args)

        if status_code == 429:
            raise exception.ImageHostRateLimitFailure(image_ref=request.url)

        request.raise_for_status()

    @staticmethod
    def check_redirect_trusted(request_response, request_session,
                               stream=True, timeout=60):
        """Check if we've been redirected to a trusted source

        Because we may be using auth, we may not want to leak authentication
        keys to an untrusted source. If we get a redirect, we need to check
        that the redirect url is one of our sources that we trust. Otherwise
        we drop the Authorization header from the redirect request. We'll
        add the header back into the request session after performing the
        request to ensure that future usage of the session.

        :param: request_response: Response object of the request to check
        :param: request_session: Session to use when redirecting
        :param: stream: Should we stream the response of the redirect
        :param: timeout: Timeout for the redirect request
        """
        # we're not a redirect, just return the original response
        if not (request_response.status_code >= 300
                and request_response.status_code < 400):
            return request_response
        # parse the destination location
        redir_url = parse.urlparse(request_response.headers['Location'])
        # close the response since we're going to replace it
        request_response.close()
        auth_header = request_session.headers.pop('Authorization', None)
        # ok we got a redirect, let's check where we are going
        secure_cdn = CONF.oci.secure_cdn_registries
        # TODO(TheJulia): Consider breaking the session calls below into
        # a helper method, because as-is, this is basically impossible
        # to unit test the delienation in behavior.
        if len([h for h in secure_cdn if h in redir_url.netloc]) > 0:
            # we're going to a trusted location, add the header back and
            # return response
            request_session.headers.update({'Authorization': auth_header})
            request_response = request_session.get(redir_url.geturl(),
                                                   stream=stream,
                                                   timeout=timeout)
        else:
            # we didn't trust the place we're going, request without auth but
            # add the auth back to the request session afterwards
            request_response = request_session.get(redir_url.geturl(),
                                                   stream=stream,
                                                   timeout=timeout)
            request_session.headers.update({'Authorization': auth_header})

        request_response.encoding = 'utf-8'
        # recheck status here to make sure we didn't get a 401 from
        # our redirect host path.
        RegistrySessionHelper.check_status(session=request_session,
                                           request=request_response)
        return request_response

    def get_token_from_config(fqdn):
        """Takes a FQDN for a container registry and consults auth config.

        This method evaluates named configuration parameter
        [oci]authentication_config and looks for pre-shared secrets
        in the supplied json file. It is written to defensively
        handle the file such that errors are not treated as fatal to
        the overall lookup process, but errors are logged.

        The expected file format is along the lines of:

        {
          "auths": {
            "domain.name": {
              "auth": "pre-shared-secret-value"
            }
          }
        }

        :param fqdn: A fully qualified domain name for interacting
                     with the remote image registry.
        :returns: String value for the "auth" key which matches
                  the supplied FQDN.
        """
        if not CONF.oci.authentication_config:
            return

        auth = None
        try:
            with open(CONF.oci.authentication_config, 'r') as auth_file:
                auth_dict = json.load(auth_file)
        except OSError as e:
            LOG.error('Failed to load pre-shared authentication token '
                      'data: %s', e)
            return
        except (json.JSONDecodeError, UnicodeDecodeError) as e:
            LOG.error('Unable to decode pre-shared authentication token '
                      'data: %s', e)
            return
        try:
            # Limiting all key interactions here to capture any formatting
            # errors in one place.
            auth_dict = auth_dict['auths']
            fqdn_dict = auth_dict.get(fqdn)
            auth = fqdn_dict.get('auth')
        except (AttributeError, KeyError):
            LOG.error('There was an error while looking up authentication '
                      'for dns name %s. Possible misformatted file?')
            return

        return auth

    @staticmethod
    def get_bearer_token(session, username=None, password=None,
                         realm=None, service=None, scope=None):
        auth = None
        token_param = {}
        if service:
            token_param['service'] = service
        if scope:
            token_param['scope'] = scope
        if username:
            # NOTE(TheJulia): This won't be invoked under the current
            # client code which does not use a username. Tokens
            # have the username encoded within and the remote servers
            # know how to decode it.
            auth = requests.auth.HTTPBasicAuth(username, password)
        elif password:
            # This is a case where we have a pre-shared token.
            LOG.debug('Using user provided pre-shared authentication '
                      'token to authenticate to the remote registry.')
            auth = requests.auth.HTTPBasicAuth('', password)
        else:
            realm_url = parse.urlparse(realm)
            local_token = RegistrySessionHelper.get_token_from_config(
                realm_url.netloc)
            if local_token:
                LOG.debug('Using a locally configured pre-shared key '
                          'for authentication to the remote registry.')
                auth = requests.auth.HTTPBasicAuth('', local_token)

        auth_req = session.get(realm, params=token_param, auth=auth,
                               timeout=CONF.webserver_connection_timeout)
        auth_req.raise_for_status()
        resp = auth_req.json()
        if 'token' not in resp:
            raise AttributeError('Invalid auth response, no token provide')
        return resp['token']

    @staticmethod
    def parse_www_authenticate(header):
        auth_type = None
        auth_type_match = re.search('^([A-Za-z]*) ', header)
        if auth_type_match:
            auth_type = auth_type_match.group(1)
        if not auth_type:
            return (None, None, None)
        realm = None
        service = None
        if 'realm=' in header:
            realm = re.search('realm="(.*?)"', header).group(1)
        if 'service=' in header:
            service = re.search('service="(.*?)"', header).group(1)
        return (auth_type, realm, service)

    @staticmethod
    @tenacity.retry(  # Retry up to 5 times with longer time for rate limit
        reraise=True,
        retry=tenacity.retry_if_exception_type(
            exception.ImageHostRateLimitFailure
        ),
        wait=tenacity.wait_random_exponential(multiplier=1.5, max=60),
        stop=tenacity.stop_after_attempt(5)
    )
    def _action(action, request_session, *args, **kwargs):
        """Perform a session action and retry if auth fails

        This function dynamically performs a specific type of call
        using the provided session (get, patch, post, etc). It will
        attempt a single re-authentication if the initial request
        fails with a 401.
        """
        _action = getattr(request_session, action)
        try:
            req = _action(*args, **kwargs)
            if not kwargs.get('stream'):
                # The caller has requested a stream, likely download so
                # we really can't call check_status because it would force
                # full content transfer.
                RegistrySessionHelper.check_status(session=request_session,
                                                   request=req)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                req = _action(*args, **kwargs)
                RegistrySessionHelper.check_status(session=request_session,
                                                   request=req)
            else:
                raise
        return req

    @staticmethod
    def get(request_session, *args, **kwargs):
        """Perform a get and retry if auth fails

        This function is designed to be used when we perform a get to
        an authenticated source. This function will attempt a single
        re-authentication request if the first one fails.
        """
        return RegistrySessionHelper._action('get',
                                             request_session,
                                             *args,
                                             **kwargs)


class OciClient(object):

    # The cached client authorization which may be used by for an
    # artifact being accessed by ironic-python-agent so we can retrieve
    # the authorization data and convey it to IPA without needing to
    # directly handle credentials to IPA.
    _cached_auth = None

    def __init__(self, verify):
        """Initialize the OCI container registry client class.

        :param verify: If certificate verification should be leveraged for
                       the underlying HTTP client.
        """
        # FIXME(TheJulia): This should come from configuration
        self.session = MakeSession(verify=verify).create()

    def authenticate(self, image_url, username=None, password=None):
        """Authenticate to the remote container registry.

        :param image_url: The URL to utilise for the remote container
                          registry.
        :param username: The username paraemter.
        :param password: The password parameter.

        :raises: AttributeError when an unknown authentication attribute has
                 been specified by the remote service.
        :raises: ImageServiceAuthenticationRequired when the remote Container
                 registry requires authentication but we do not have a
                 credentials.
        """
        url = self._image_to_url(image_url)
        image, tag = self._image_tag_from_url(url)
        scope = 'repository:%s:pull' % image[1:]

        url = self._build_url(url, path='/')

        # If authenticate is called an additional time....
        # clear the authorization in the client.
        if self.session:
            self.session.headers.pop('Authorization', None)

        r = self.session.get(url, timeout=CONF.webserver_connection_timeout)
        LOG.debug('%s status code %s', url, r.status_code)
        if r.status_code == 200:
            # "Auth" was successful, returning.
            return self.session
        if r.status_code != 401:
            # Auth was rejected.
            r.raise_for_status()
        if 'www-authenticate' not in r.headers:
            # Something is wrong and unexpected.
            raise AttributeError(
                'Unknown authentication method for headers: %s' % r.headers)

        auth = None
        www_auth = r.headers['www-authenticate']
        token_param = {}
        (auth_type, realm, service) = \
            RegistrySessionHelper.parse_www_authenticate(www_auth)

        if auth_type and auth_type.lower() == 'bearer':
            LOG.debug('Using bearer token auth')
            token = RegistrySessionHelper.get_bearer_token(
                self.session,
                username=username,
                password=password,
                realm=realm,
                service=service,
                scope=scope)
        elif auth_type and auth_type.lower() == 'basic':
            if not username or not password:
                raise exception.ImageServiceAuthenticationRequired(
                    image_ref=image_url)
            auth = requests_auth.HTTPBasicAuth(username, password)
            rauth = self.session.get(
                url, params=token_param,
                auth=auth,
                timeout=CONF.webserver_connection_timeout)
            rauth.raise_for_status()
            token = (
                base64.b64encode(
                    bytes(username + ':' + password, 'utf-8')).decode('ascii')
            )
        else:
            raise AttributeError(
                'Unknown www-authenticate value: %s', www_auth)
        auth_header = '%s %s' % (auth_type, token)
        self.session.headers['Authorization'] = auth_header
        # Set a cached Authorization token value so we can extract it
        # if needed, useful for enabling something else to be able to
        # make that actual call.
        self._cached_auth = auth_header
        setattr(self.session, 'reauthenticate', self.authenticate)
        setattr(
            self.session,
            'auth_args',
            dict(
                image_url=image_url,
                username=username,
                password=password,
                session=self.session
            )
        )

    @staticmethod
    def _get_response_text(response, encoding='utf-8', force_encoding=False):
        """Return request response text

        We need to set the encoding for the response other wise it
        will attempt to detect the encoding which is very time consuming.
        See https://github.com/psf/requests/issues/4235 for additional
        context.

        :param: response: requests Respoinse object
        :param: encoding: encoding to set if not currently set
        :param: force_encoding: set response encoding always
        """

        if force_encoding or not response.encoding:
            response.encoding = encoding
        return response.text

    @staticmethod
    def _get_response_json(response, encoding='utf-8', force_encoding=False,
                           validate_digest=None):
        """Return request response as JSON.

        We need to set the encoding for the response other wise it
        will attempt to detect the encoding which is very time consuming.
        See https://github.com/psf/requests/issues/4235 for additional
        context.

        The docker-content-digest header is added as a dockerContentDigest
        field to the response.

        :param: response: requests Respoinse object
        :param: encoding: encoding to set if not currently set
        :param: force_encoding: set response encoding always
        :param: validate_digest: checksum to validate the text against
        """
        text = OciClient._get_response_text(response, encoding, force_encoding)
        if validate_digest:
            checksum_utils.validate_text_checksum(text, validate_digest)
        resource = json.loads(text)
        contentDigest = response.headers.get('docker-content-digest')
        if contentDigest:
            resource['dockerContentDigest'] = contentDigest
        return resource

    @classmethod
    def _build_url(cls, url, path):
        """Build an HTTPS URL from the input urlparse data.

        :param url: The urlparse result object with the netloc object which
                    is extracted and used by this method.
        :param path: The path in the form of a string which is then assembled
                     into an HTTPS URL to be used for access.
        :returns: A fully formed url in the form of https://ur.
        """
        netloc = url.netloc
        scheme = 'https'
        return '%s://%s/v2%s' % (scheme, netloc, path)

    def _get_manifest(self, image_url, digest=None):

        if not digest:
            # Caller has the digest in the url, that's fine, lets
            # use that.
            digest = image_url.path.split('@')[1]
        image_path = image_url.path.split(':')[0]

        manifest_url = self._build_url(
            image_url, CALL_MANIFEST % {'image': image_path,
                                        'tag': digest})

        # Explicitly ask for the OCI artifact index
        manifest_headers = {'Accept': ", ".join([MEDIA_OCI_MANIFEST_V1])}
        try:
            manifest_r = RegistrySessionHelper.get(
                self.session,
                manifest_url,
                headers=manifest_headers,
                timeout=CONF.webserver_connection_timeout
            )
        except requests.exceptions.HTTPError as e:
            LOG.error('Encountered error while attempting to download '
                      'manifest %s for image %s: %s',
                      manifest_url, image_url, e)
            if e.response.status_code == 401:
                # Authorization Required.
                raise exception.ImageServiceAuthenticationRequired(
                    image_ref=manifest_url)
            if e.response.status_code in (403, 404):
                raise exception.ImageNotFound(
                    image_id=image_url.geturl())
            if e.response.status_code >= 500:
                raise exception.TemporaryFailure()
            raise
        return self._get_response_json(
            manifest_r, validate_digest=digest)

    def _get_artifact_index(self, image_url):
        LOG.debug('Attempting to get the artifact index for: %s',
                  image_url)
        parts = self._resolve_tag(image_url)
        index_url = self._build_url(
            image_url, CALL_MANIFEST % parts
        )
        # Explicitly ask for the OCI artifact index, fall back to manifest
        index_headers = {'Accept': ", ".join([MEDIA_OCI_INDEX_V1,
                                              MEDIA_OCI_MANIFEST_V1])}

        try:
            index_r = RegistrySessionHelper.get(
                self.session,
                index_url,
                headers=index_headers,
                timeout=CONF.webserver_connection_timeout
            )
        except requests.exceptions.HTTPError as e:
            LOG.error('Encountered error while attempting to download '
                      'artifact index %s for image %s: %s',
                      index_url, image_url, e)
            if e.response.status_code == 401:
                # Authorization Required.
                raise exception.ImageServiceAuthenticationRequired(
                    image_ref=index_url)
            if e.response.status_code in (403, 404):
                raise exception.ImageNotFound(
                    image_id=image_url.geturl())
            if e.response.status_code >= 500:
                raise exception.TemporaryFailure()
            raise
        # Return a dictionary to the caller so it can house the
        # filtering/sorting application logic.
        return self._get_response_json(index_r)

    def _resolve_tag(self, image_url):
        """Attempts to resolve tags from a container URL."""
        LOG.debug('Attempting to resolve tag for: %s',
                  image_url)
        image, tag = self._image_tag_from_url(image_url)
        parts = {
            'image': image,
            'tag': tag
        }
        tags_url = self._build_url(
            image_url, CALL_TAGS % parts
        )
        tag_headers = {'Accept': ", ".join([MEDIA_OCI_INDEX_V1])}
        try:
            tags_r = RegistrySessionHelper.get(
                self.session, tags_url,
                headers=tag_headers,
                timeout=CONF.webserver_connection_timeout)
        except requests.exceptions.HTTPError as e:
            LOG.error('Encountered error while attempting to download '
                      'the tag list %s for image %s: %s',
                      tags_url, image_url, e)
            if e.response.status_code == 401:
                # Authorization Required.
                raise exception.ImageServiceAuthenticationRequired(
                    image_ref=tags_url)
            if e.response.status_code >= 500:
                raise exception.TemporaryFailure()
            raise
        tags = tags_r.json()['tags']
        while 'next' in tags_r.links:
            next_url = parse.urljoin(tags_url, tags_r.links['next']['url'])
            try:
                tags_r = RegistrySessionHelper.get(
                    self.session, next_url,
                    headers=tag_headers,
                    timeout=CONF.webserver_connection_timeout)
            except requests.exceptions.HTTPError as e:
                LOG.error('Encountered error while attempting to download '
                          'the next tag list %s for image %s: %s',
                          next_url, image_url, e)
                raise
            tags.extend(tags_r.json()['tags'])
        if not tags:
            raise exception.InvalidImageRef(
                _("Image %s does not have any tags") % image_url.geturl())
        if tag not in tags:
            raise exception.OciImageTagNotFound(
                image_url=image_url.geturl(),
                tags=', '.join(tags))
        return parts

    def get_artifact_index(self, image):
        """Retrieve an index of artifacts in the Container Registry.

        :param image: The remote container registry URL in the form of
                      oci://host/user/container:tag.

        :returns: A dictionary object representing the index of artifacts
                  present in the container registry, in the form of manifest
                  references along with any other metadata per entry which
                  the remote registry returns such as annotations, and
                  platform labeling which aids in artifact selection.
        """
        image_url = self._image_to_url(image)
        return self._get_artifact_index(image_url)

    def get_manifest(self, image, digest=None):
        """Retrieve a manifest from the remote API.

        This method is a wrapper for the _get_manifest helper, which
        normalizes the input URL, performs basic sanity checking,
        and then calls the underlying method to retrieve the manifest.

        The manifest is then returned to the caller in the form of a
        dictionary.

        :param image: The full URL to the desired manifest or the URL
                      of the container and an accompanying digest parameter.
        :param digest: The Digest value for the requested manifest.
        :returns: A dictionary object representing the manifest as stored
                  in the remote API.
        """
        LOG.debug('Attempting to get manifest for: %s', image)
        if not digest and '@' in image:
            # Digest must be part of the URL, this is fine!
            url_split = image.split("@")
            image_url = self._image_to_url(url_split[0])
            digest = url_split[1]
        elif digest and '@' in image:
            raise AttributeError('Invalid request - Appears to attempt '
                                 'to use a digest value and a digest in '
                                 'the provided URL.')
        else:
            image_url = self._image_to_url(image)
        return self._get_manifest(image_url, digest)

    def get_blob_url(self, image, blob_digest):
        """Generates an HTTP representing an blob artifact.

        :param image: The OCI Container URL.
        :param blob_digest: The digest value representing the desired blob
                            artifact.
        :returns: A HTTP URL string representing the blob URL which can be
                  utilized by an HTTP client to retrieve the artifact.
        """
        if not blob_digest and '@' in image:
            split_url = image.split('@')
            image_url = split_url[0]
            blob_digest = split_url[1]
        elif blob_digest and '@' in image:
            split_url = image.split('@')
            image_url = split_url[0]
            # The caller likely has a bug or bad pattern
            # which needs to be fixed
        else:
            image_url = image
        image_url = self._image_to_url(image_url)
        # just in caes, split out the tag since it is not
        # used for a blob manifest lookup.
        image_path = image_url.path.split(':')[0]
        manifest_url = self._build_url(
            image_url, CALL_BLOB % {'image': image_path,
                                    'digest': blob_digest})
        return manifest_url

    def get_cached_auth(self):
        """Retrieves the cached authentication header for reuse."""
        # This enables the cached authentication data to be retrieved
        # to enable Ironic to provide that the data without shipping
        # credentials around directly.
        return self._cached_auth

    def download_blob_from_manifest(self, manifest_url, image_file):
        """Retrieves the requested blob from the manifest URL...

        And saves the requested manifest's artifact as the requested
        image_file location, and then returns the verified checksum.

        :param manifest_url: The URL, in oci://host/user/container@digest
                             formatted artifact manifest URL. This is *not*
                             the digest value for the blob, which can only
                             be discovered by retrieving the manifest.
        :param image_file: The image file object to write the blob to.
        :returns: The verified digest value matching the saved artifact.
        """
        LOG.debug('Starting download blob download sequence for %s',
                  manifest_url)
        manifest = self.get_manifest(manifest_url)
        layers = manifest.get('layers', [])
        layer_count = len(layers)
        if layer_count != 1:
            # This is not a blob manifest, it is the container,
            # or something else we don't understand.
            raise exception.ImageRefValidationFailed(
                'Incorrect number of layers. Expected 1 layer, '
                'found %s layers.' % layer_count)
        blob_digest = layers[0].get('digest')
        blob_url = self.get_blob_url(manifest_url, blob_digest)
        LOG.debug('Identified download url for blob: %s', blob_url)
        # One which is an OCI URL with a manifest.
        try:
            resp = RegistrySessionHelper.get(
                self.session,
                blob_url,
                stream=True,
                timeout=CONF.webserver_connection_timeout
            )
            resp = RegistrySessionHelper.check_redirect_trusted(
                resp, self.session, stream=True)
            if resp.status_code != 200:
                raise exception.ImageRefValidationFailed(
                    image_href=blob_url,
                    reason=("Got HTTP code %s instead of 200 in response "
                            "to GET request.") % resp.status_code)
            # Reminder: image_file, is a file object handler.
            split_digest = blob_digest.split(':')

            # Invoke the transfer helper so the checksum can be calculated
            # in transfer.
            download_helper = checksum_utils.TransferHelper(
                resp, split_digest[0], split_digest[1])
            # NOTE(TheJuila): If we *ever* try to have retry logic here,
            # remember to image_file.seek(0) to reset position.
            for chunk in download_helper:
                # write the desired file out
                image_file.write(chunk)
            LOG.debug('Download of %(manifest)s has completed. Transferred '
                      '%(bytes)s of %(total)s total bytes.',
                      {'manifest': manifest_url,
                       'bytes': download_helper.bytes_transferred,
                       'total': download_helper.content_length})
            if download_helper.checksum_matches:
                return blob_digest
            else:
                raise exception.ImageChecksumError()

        except requests.exceptions.HTTPError as e:
            LOG.error('Encountered error while attempting to download %s: %s',
                      blob_url, e)
            # Stream changes the behavior, so odds of hitting
            # this area area a bit low unless an actual exception
            # is raised.
            if e.response.status_code == 401:
                # Authorization Required.
                raise exception.ImageServiceAuthenticationRequired(
                    image_ref=blob_url)
            if e.response.status_code in (403, 404):
                raise exception.ImageNotFound(image_id=blob_url)
            if e.response.status_code >= 500:
                raise exception.TemporaryFailure()
            raise

        except (OSError, requests.ConnectionError, requests.RequestException,
                IOError) as e:
            LOG.error('Encountered error while attempting to download %s: %s',
                      blob_url, e)
            raise exception.ImageDownloadFailed(image_href=blob_url,
                                                reason=str(e))

    @classmethod
    def _image_tag_from_url(cls, image_url):
        """Identify image and tag from image_url.

        :param image_url: Input image url.
        :returns: Tuple of values, image URL which has been reconstructed
                  and the requested tag. Defaults to 'latest' when a tag has
                  not been identified as part of the supplied URL.
        """
        if '@' in image_url.path:
            parts = image_url.path.split('@')
            tag = parts[-1]
            image = ':'.join(parts[:-1])
        elif ':' in image_url.path:
            parts = image_url.path.split(':')
            tag = parts[-1]
            image = ':'.join(parts[:-1])
        else:
            tag = 'latest'
            image = image_url.path
        return image, tag

    @classmethod
    def _image_to_url(cls, image):
        """Helper to create an OCI URL."""
        if '://' not in image:
            # Slight bit of future proofing in case we ever support
            # identifying bare URLs.
            image = 'oci://' + image
        url = parse.urlparse(image)
        return url



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import base64
import requests
import tenacity

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import swift

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


# NOTE(TheJulia): Deprecated after the 2024.1 PTG in favor of
# a future step templating interface.

def save_configuration(task, url, data):
    """Store configuration mold to indicated location.

    :param task: A TaskManager instance.
    :param name: URL of the configuration item to save to.
    :param data: Content of JSON data to save.

    :raises IronicException: If using Swift storage and no authentication
        token found in task's context.
    :raises HTTPError: If failed to complete HTTP request.
    """
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            requests.exceptions.ConnectionError),
        stop=tenacity.stop_after_attempt(CONF.molds.retry_attempts),
        wait=tenacity.wait_fixed(CONF.molds.retry_interval),
        reraise=True
    )
    def _request(url, data, auth_header):
        return requests.put(
            url, data=json.dumps(data, indent=2), headers=auth_header,
            timeout=CONF.webserver_connection_timeout)

    auth_header = _get_auth_header(task)
    response = _request(url, data, auth_header)
    response.raise_for_status()


# NOTE(TheJulia): Deprecated after the 2024.1 PTG in favor of
# a future step templating interface.

def get_configuration(task, url):
    """Gets configuration mold from indicated location.

    :param task: A TaskManager instance.
    :param url: URL of the configuration item to get.

    :returns: JSON configuration mold

    :raises IronicException: If using Swift storage and no authentication
        token found in task's context.
    :raises HTTPError: If failed to complete HTTP request.
    """
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            requests.exceptions.ConnectionError),
        stop=tenacity.stop_after_attempt(CONF.molds.retry_attempts),
        wait=tenacity.wait_fixed(CONF.molds.retry_interval),
        reraise=True
    )
    def _request(url, auth_header):
        return requests.get(url, headers=auth_header,
                            timeout=CONF.webserver_connection_timeout)

    auth_header = _get_auth_header(task)
    response = _request(url, auth_header)
    if response.status_code == requests.codes.ok:
        if not response.content:
            raise exception.IronicException(_(
                "Configuration mold for node %(node_uuid)s at %(url)s is "
                "empty") % {'node_uuid': task.node.uuid, 'url': url})
        try:
            return response.json()
        except json.decoder.JSONDecodeError as jde:
            raise exception.IronicException(_(
                "Configuration mold for node %(node_uuid)s at %(url)s has "
                "invalid JSON: %(error)s)")
                % {'node_uuid': task.node.uuid, 'url': url, 'error': jde})

    response.raise_for_status()


# NOTE(TheJulia): Deprecated after the 2024.1 PTG in favor of
# a future step templating interface.

def _get_auth_header(task):
    """Based on setup of configuration mold storage gets authentication header

    :param task: A TaskManager instance.
    :raises IronicException: If using Swift storage and no authentication
        token found in task's context.
    """
    auth_header = None
    if CONF.molds.storage == 'swift':
        # TODO(ajya) Need to update to use Swift client and context session
        auth_token = swift.get_swift_session().get_token()
        if auth_token:
            auth_header = {'X-Auth-Token': auth_token}
        else:
            raise exception.IronicException(
                _('Missing auth_token for configuration mold access for node '
                  '%s') % task.node.uuid)
    elif CONF.molds.storage == 'http':
        if CONF.molds.user and CONF.molds.password:
            auth_header = {'Authorization': 'Basic %s'
                           % base64.encode_as_text(
                               '%s:%s' % (CONF.molds.user,
                                          CONF.molds.password))}
    return auth_header



# Copyright 2016 Mirantis Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from ironic.conf import auth as ironic_auth
from ironic.tests import base


class AuthConfTestCase(base.TestCase):

    def setUp(self):
        super(AuthConfTestCase, self).setUp()
        self.test_group = 'test_group'
        self.cfg_fixture.conf.register_group(cfg.OptGroup(self.test_group))
        ironic_auth.register_auth_opts(self.cfg_fixture.conf, self.test_group)
        self.config(auth_type='password',
                    group=self.test_group)
        # NOTE(pas-ha) this is due to auth_plugin options
        # being dynamically registered on first load,
        # but we need to set the config before
        plugin = ks_loading.get_plugin_loader('password')
        opts = ks_loading.get_auth_plugin_conf_options(plugin)
        self.cfg_fixture.register_opts(opts, group=self.test_group)
        self.config(auth_url='http://127.0.0.1:9898',
                    username='fake_user',
                    password='fake_pass',
                    project_name='fake_tenant',
                    group=self.test_group)

    def test_add_auth_opts(self):
        opts = ironic_auth.add_auth_opts([])
        # check that there is no duplicates
        names = {o.dest for o in opts}
        self.assertEqual(len(names), len(opts))
        # NOTE(pas-ha) checking for most standard auth and session ones only
        expected = {'timeout', 'insecure', 'cafile', 'certfile', 'keyfile',
                    'auth_type', 'auth_url', 'username', 'password',
                    'tenant_name', 'project_name', 'trust_id',
                    'domain_id', 'user_domain_id', 'project_domain_id'}
        self.assertTrue(expected.issubset(names))



# Copyright 2016 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics
from ironic.common import metrics_collector
from ironic.common import metrics_statsd


CONF = cfg.CONF


def get_metrics_logger(prefix='', backend=None, host=None, delimiter='.'):
    """Return a metric logger with the specified prefix.

    The format of the prefix is:
    [global_prefix<delim>][host_name<delim>]prefix
    where <delim> is the delimiter (default is '.')

    :param prefix: Prefix for this metric logger.
        Value should be a string or None.
    :param backend: Backend to use for the metrics system.
        Possible values are 'noop' and 'statsd'.
    :param host: Name of this node.
    :param delimiter: Delimiter to use for the metrics name.
    :return: The new MetricLogger.
    """
    if not isinstance(prefix, str):
        msg = (_("This metric prefix (%s) is of unsupported type. "
                 "Value should be a string or None")
               % str(prefix))
        raise exception.InvalidMetricConfig(msg)

    if CONF.metrics.prepend_host and host:
        if CONF.metrics.prepend_host_reverse:
            host = '.'.join(reversed(host.split('.')))

        if prefix:
            prefix = delimiter.join([host, prefix])
        else:
            prefix = host

    if CONF.metrics.global_prefix:
        if prefix:
            prefix = delimiter.join([CONF.metrics.global_prefix, prefix])
        else:
            prefix = CONF.metrics.global_prefix

    backend = backend or CONF.metrics.backend
    if backend == 'statsd':
        return metrics_statsd.StatsdMetricLogger(prefix, delimiter=delimiter)
    elif backend == 'noop':
        return metrics.NoopMetricLogger(prefix, delimiter=delimiter)
    elif backend == 'collector':
        return metrics_collector.DictCollectionMetricLogger(
            prefix, delimiter=delimiter)
    else:
        msg = (_("The backend is set to an unsupported type: "
                 "%s. Value should be 'noop' or 'statsd'.")
               % backend)
        raise exception.InvalidMetricConfig(msg)



#    Copyright (C) 2025 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import io
import json
from unittest import mock
from urllib import parse

from oslo_config import cfg
import requests

from ironic.common import exception
from ironic.common import oci_registry
from ironic.tests import base

CONF = cfg.CONF


ACCEPT_INDEX_OR_MANIFEST = (
    'application/vnd.oci.image.index.v1+json, '
    'application/vnd.oci.image.manifest.v1+json'
)


class OciClientTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.client = oci_registry.OciClient(verify=False)

    @mock.patch.object(oci_registry, 'MakeSession',
                       autospec=True)
    def test_client_init_make_session(self, mock_session):
        oci_registry.OciClient(verify=True)
        mock_session.assert_called_once_with(verify=True)
        mock_session.return_value.create.assert_called_once()

    def test__image_to_url(self):
        url = self.client._image_to_url('oci://host/path')
        self.assertEqual('host', url.netloc)
        self.assertEqual('/path', url.path)
        self.assertEqual('oci', url.scheme)

    def test__image_to_url_adds_oci(self):
        url = self.client._image_to_url('host/path')
        self.assertEqual('oci', url.scheme)
        self.assertEqual('host', url.netloc)
        self.assertEqual('/path', url.path)

    def test_image_tag_from_url(self):
        url = self.client._image_to_url('oci://host/path')
        img, tag = self.client._image_tag_from_url(url)
        self.assertEqual('/path', img)
        self.assertEqual('latest', tag)

    def test_image_tag_from_url_with_tag(self):
        url = self.client._image_to_url('oci://host/path:5.3')
        img, tag = self.client._image_tag_from_url(url)
        self.assertEqual('/path', img)
        self.assertEqual('5.3', tag)

    def test_image_tag_from_url_with_digest(self):
        url = self.client._image_to_url('oci://host/path@sha256:f00')
        img, tag = self.client._image_tag_from_url(url)
        self.assertEqual('/path', img)
        self.assertEqual('sha256:f00', tag)

    def test_get_blob_url(self):
        digest = ('sha256:' + 'a' * 64)
        image = 'oci://host/project/container'
        res = self.client.get_blob_url(image, digest)
        self.assertEqual(
            'https://host/v2/project/container/blobs/' + digest,
            res)


@mock.patch.object(requests.sessions.Session, 'get', autospec=True)
class OciClientRequestTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.client = oci_registry.OciClient(verify=True)

    def test_get_manifest_checksum_verifies(self, get_mock):
        fake_csum = 'f' * 64
        get_mock.return_value.status_code = 200
        get_mock.return_value.text = '{}'
        self.assertRaises(
            exception.ImageChecksumError,
            self.client.get_manifest,
            'oci://localhost/local@sha256:' + fake_csum)
        get_mock.return_value.assert_has_calls([
            mock.call.raise_for_status(),
            mock.call.encoding.__bool__()])
        get_mock.assert_called_once_with(
            mock.ANY,
            ('https://localhost/v2/local/manifests/sha256:ffffffffff'
             'ffffffffffffffffffffffffffffffffffffffffffffffffffffff'),
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)

    def test_get_manifest(self, get_mock):
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_mock.return_value.status_code = 200
        get_mock.return_value.text = '{}'
        get_mock.return_value.headers = {}
        res = self.client.get_manifest(
            'oci://localhost/local@sha256:' + csum)
        self.assertEqual({}, res)
        get_mock.return_value.assert_has_calls([
            mock.call.raise_for_status(),
            mock.call.encoding.__bool__()])
        get_mock.assert_called_once_with(
            mock.ANY,
            'https://localhost/v2/local/manifests/sha256:' + csum,
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)

    def test_get_manifest_with_content_digest(self, get_mock):
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_mock.return_value.status_code = 200
        get_mock.return_value.text = '{}'
        get_mock.return_value.headers = {'docker-content-digest': 'abcd'}
        res = self.client.get_manifest(
            'oci://localhost/local@sha256:' + csum)
        self.assertEqual({'dockerContentDigest': 'abcd'}, res)
        get_mock.return_value.assert_has_calls([
            mock.call.raise_for_status(),
            mock.call.encoding.__bool__()])
        get_mock.assert_called_once_with(
            mock.ANY,
            'https://localhost/v2/local/manifests/sha256:' + csum,
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)

    def test_get_manifest_auth_required(self, get_mock):
        fake_csum = 'f' * 64
        response = mock.Mock()
        response.status_code = 401
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageServiceAuthenticationRequired,
            self.client.get_manifest,
            'oci://localhost/local@sha256:' + fake_csum)
        call_mock = mock.call(
            mock.ANY,
            ('https://localhost/v2/local/manifests/sha256:ffffffffff'
             'ffffffffffffffffffffffffffffffffffffffffffffffffffffff'),
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)
        # Gets retried.
        get_mock.assert_has_calls([call_mock, call_mock])

    def test_get_manifest_image_access_denied(self, get_mock):
        fake_csum = 'f' * 64
        response = mock.Mock()
        response.status_code = 403
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageNotFound,
            self.client.get_manifest,
            'oci://localhost/local@sha256:' + fake_csum)
        call_mock = mock.call(
            mock.ANY,
            ('https://localhost/v2/local/manifests/sha256:ffffffffff'
             'ffffffffffffffffffffffffffffffffffffffffffffffffffffff'),
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test_get_manifest_image_not_found(self, get_mock):
        fake_csum = 'f' * 64
        response = mock.Mock()
        response.status_code = 404
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageNotFound,
            self.client.get_manifest,
            'oci://localhost/local@sha256:' + fake_csum)
        call_mock = mock.call(
            mock.ANY,
            ('https://localhost/v2/local/manifests/sha256:ffffffffff'
             'ffffffffffffffffffffffffffffffffffffffffffffffffffffff'),
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test_get_manifest_image_temporary_failure(self, get_mock):
        fake_csum = 'f' * 64
        response = mock.Mock()
        response.status_code = 500
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.TemporaryFailure,
            self.client.get_manifest,
            'oci://localhost/local@sha256:' + fake_csum)
        call_mock = mock.call(
            mock.ANY,
            ('https://localhost/v2/local/manifests/sha256:ffffffffff'
             'ffffffffffffffffffffffffffffffffffffffffffffffffffffff'),
            headers={'Accept': 'application/vnd.oci.image.manifest.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    @mock.patch.object(oci_registry.OciClient, '_resolve_tag',
                       autospec=True)
    def test_get_artifact_index_with_tag(self, resolve_tag_mock, get_mock):
        resolve_tag_mock.return_value = {
            'image': '/local',
            'tag': 'tag'
        }
        get_mock.return_value.status_code = 200
        get_mock.return_value.text = '{}'
        get_mock.return_value.headers = {}
        res = self.client.get_artifact_index(
            'oci://localhost/local:tag')
        self.assertEqual({}, res)
        resolve_tag_mock.assert_called_once_with(
            mock.ANY,
            parse.urlparse('oci://localhost/local:tag'))
        get_mock.return_value.assert_has_calls([
            mock.call.raise_for_status(),
            mock.call.encoding.__bool__()])
        get_mock.assert_called_once_with(
            mock.ANY,
            'https://localhost/v2/local/manifests/tag',
            headers={'Accept': ACCEPT_INDEX_OR_MANIFEST},
            timeout=60)

    @mock.patch.object(oci_registry.OciClient, '_resolve_tag',
                       autospec=True)
    def test_get_artifact_index_not_found(self, resolve_tag_mock, get_mock):
        resolve_tag_mock.return_value = {
            'image': '/local',
            'tag': 'tag'
        }
        response = mock.Mock()
        response.status_code = 404
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageNotFound,
            self.client.get_artifact_index,
            'oci://localhost/local:tag')
        resolve_tag_mock.assert_called_once_with(
            mock.ANY,
            parse.urlparse('oci://localhost/local:tag'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/manifests/tag',
            headers={'Accept': ACCEPT_INDEX_OR_MANIFEST},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    @mock.patch.object(oci_registry.OciClient, '_resolve_tag',
                       autospec=True)
    def test_get_artifact_index_not_authorized(self, resolve_tag_mock,
                                               get_mock):
        resolve_tag_mock.return_value = {
            'image': '/local',
            'tag': 'tag'
        }
        response = mock.Mock()
        response.status_code = 401
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageServiceAuthenticationRequired,
            self.client.get_artifact_index,
            'oci://localhost/local:tag')
        resolve_tag_mock.assert_called_once_with(
            mock.ANY,
            parse.urlparse('oci://localhost/local:tag'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/manifests/tag',
            headers={'Accept': ACCEPT_INDEX_OR_MANIFEST},
            timeout=60)
        # Automatic retry to authenticate
        get_mock.assert_has_calls([call_mock, call_mock])

    @mock.patch.object(oci_registry.OciClient, '_resolve_tag',
                       autospec=True)
    def test_get_artifact_index_temporaryfailure(self, resolve_tag_mock,
                                                 get_mock):
        resolve_tag_mock.return_value = {
            'image': '/local',
            'tag': 'tag'
        }
        response = mock.Mock()
        response.status_code = 500
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.TemporaryFailure,
            self.client.get_artifact_index,
            'oci://localhost/local:tag')
        resolve_tag_mock.assert_called_once_with(
            mock.ANY,
            parse.urlparse('oci://localhost/local:tag'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/manifests/tag',
            headers={'Accept': ACCEPT_INDEX_OR_MANIFEST},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    @mock.patch.object(oci_registry.OciClient, '_resolve_tag',
                       autospec=True)
    def test_get_artifact_index_access_denied(self, resolve_tag_mock,
                                              get_mock):
        resolve_tag_mock.return_value = {
            'image': '/local',
            'tag': 'tag'
        }
        response = mock.Mock()
        response.status_code = 403
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageNotFound,
            self.client.get_artifact_index,
            'oci://localhost/local:tag')
        resolve_tag_mock.assert_called_once_with(
            mock.ANY,
            parse.urlparse('oci://localhost/local:tag'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/manifests/tag',
            headers={'Accept': ACCEPT_INDEX_OR_MANIFEST},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test__resolve_tag(self, get_mock):
        response = mock.Mock()
        response.json.return_value = {'tags': ['latest', 'foo', 'bar']}
        response.status_code = 200
        response.links = {}
        get_mock.return_value = response
        res = self.client._resolve_tag(
            parse.urlparse('oci://localhost/local'))
        self.assertDictEqual({'image': '/local', 'tag': 'latest'}, res)
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test__resolve_tag_if_not_found(self, get_mock):
        response = mock.Mock()
        response.json.return_value = {'tags': ['foo', 'bar']}
        response.status_code = 200
        response.links = {}
        get_mock.return_value = response
        self.assertRaises(
            exception.OciImageTagNotFound,
            self.client._resolve_tag,
            parse.urlparse('oci://localhost/local'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test__resolve_tag_if_no_tags(self, get_mock):
        response = mock.Mock()
        response.json.return_value = {'tags': []}
        response.status_code = 200
        response.links = {}
        get_mock.return_value = response
        self.assertRaises(
            exception.InvalidImageRef,
            self.client._resolve_tag,
            parse.urlparse('oci://localhost/local'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test__resolve_tag_follows_links(self, get_mock):
        response = mock.Mock()
        response.json.return_value = {'tags': ['foo', 'bar']}
        response.status_code = 200
        response.links = {'next': {'url': 'list2'}}
        response2 = mock.Mock()
        response2.json.return_value = {'tags': ['zoo']}
        response2.status_code = 200
        response2.links = {}
        get_mock.side_effect = iter([response, response2])
        res = self.client._resolve_tag(
            parse.urlparse('oci://localhost/local:zoo'))
        self.assertDictEqual({'image': '/local', 'tag': 'zoo'}, res)
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        call_mock_2 = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list2',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock, call_mock_2])

    def test__resolve_tag_auth_needed(self, get_mock):
        response = mock.Mock()
        response.json.return_value = {}
        response.status_code = 401
        response.text = 'Authorization Required'
        response.links = {}
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.ImageServiceAuthenticationRequired,
            self.client._resolve_tag,
            parse.urlparse('oci://localhost/local'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test__resolve_tag_temp_failure(self, get_mock):
        response = mock.Mock()
        response.json.return_value = {}
        response.status_code = 500
        response.text = 'Server on vacation'
        response.links = {}
        exc = requests.exceptions.HTTPError(
            response=response)
        get_mock.side_effect = exc
        self.assertRaises(
            exception.TemporaryFailure,
            self.client._resolve_tag,
            parse.urlparse('oci://localhost/local'))
        call_mock = mock.call(
            mock.ANY,
            'https://localhost/v2/local/tags/list',
            headers={'Accept': 'application/vnd.oci.image.index.v1+json'},
            timeout=60)
        get_mock.assert_has_calls([call_mock])

    def test_authenticate_noop(self, get_mock):
        """Test authentication when the remote endpoint doesn't require it."""
        response = mock.Mock()
        response.status_code = 200
        get_mock.return_value = response
        self.client.authenticate(
            'oci://localhost/foo/bar:meow',
            username='foo',
            password='bar')
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60)])

    def test_authenticate_401_no_header(self, get_mock):
        """Test authentication when the remote endpoint doesn't require it."""
        response = mock.Mock()
        response.status_code = 401
        response.headers = {}
        get_mock.return_value = response
        self.assertRaisesRegex(
            AttributeError,
            'Unknown authentication method',
            self.client.authenticate,
            'oci://localhost/foo/bar:meow',
            username='foo',
            password='bar')
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60)])

    def test_authenticate_401_bad_header(self, get_mock):
        """Test authentication when the remote endpoint doesn't require it."""
        response = mock.Mock()
        response.status_code = 401
        response.headers = {'www-authenticate': 'magic'}
        get_mock.return_value = response
        self.assertRaisesRegex(
            AttributeError,
            'Unknown www-authenticate value',
            self.client.authenticate,
            'oci://localhost/foo/bar:meow',
            username='foo',
            password='bar')
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60)])

    def test_authenticate_401_bearer_auth(self, get_mock):
        self.assertIsNone(self.client._cached_auth)
        self.assertIsNone(self.client.session.headers.get('Authorization'))
        response = mock.Mock()
        response.status_code = 401
        response.json.return_value = {'token': 'me0w'}
        response.headers = {'www-authenticate': 'bearer realm="foo"'}
        response2 = mock.Mock()
        response2.status_code = 200
        response2.json.return_value = {'token': 'me0w'}
        get_mock.side_effect = iter([response, response2])
        self.client.authenticate(
            'oci://localhost/foo/bar:meow',
            username='',
            password='bar')
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60),
            mock.call(mock.ANY, 'foo',
                      params={'scope': 'repository:foo/bar:pull'},
                      auth=mock.ANY, timeout=60)])
        self.assertIsNotNone(self.client._cached_auth)
        self.assertEqual('bearer me0w',
                         self.client.session.headers['Authorization'])

    def test_authenticate_401_basic_auth_no_username(self, get_mock):
        self.assertIsNone(self.client._cached_auth)
        self.assertIsNone(self.client.session.headers.get('Authorization'))
        response = mock.Mock()
        response.status_code = 401
        response.headers = {'www-authenticate': 'basic service="foo"'}
        get_mock.return_value = response
        self.assertRaises(
            exception.ImageServiceAuthenticationRequired,
            self.client.authenticate,
            'oci://localhost/foo/bar:meow',
            username='',
            password='bar')
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60)])

    def test_authenticate_401_basic_auth(self, get_mock):
        self.assertIsNone(self.client._cached_auth)
        self.assertIsNone(self.client.session.headers.get('Authorization'))
        response = mock.Mock()
        response.status_code = 401
        response.headers = {'www-authenticate': 'basic service="foo"'}
        response2 = mock.Mock()
        response2.status_code = 200
        get_mock.side_effect = iter([response, response2])
        self.client.authenticate(
            'oci://localhost/foo/bar:meow',
            username='user',
            password='bar')
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60),
            mock.call(mock.ANY, 'https://localhost/v2/',
                      params={},
                      auth=mock.ANY, timeout=60)])
        self.assertIsNotNone(self.client._cached_auth)
        self.assertEqual('basic dXNlcjpiYXI=',
                         self.client.session.headers['Authorization'])

    @mock.patch.object(oci_registry.RegistrySessionHelper,
                       'get_token_from_config',
                       autospec=True)
    def test_authenticate_401_fallback_to_service_config(self, token_mock,
                                                         get_mock):
        self.assertIsNone(self.client._cached_auth)
        self.assertIsNone(self.client.session.headers.get('Authorization'))
        response = mock.Mock()
        response.status_code = 401
        response.headers = {
            'www-authenticate': 'bearer realm="https://foo/bar"'}
        response2 = mock.Mock()
        response2.status_code = 200
        response2.json.return_value = {'token': 'me0w'}
        get_mock.side_effect = iter([response, response2])
        self.client.authenticate(
            'oci://localhost/foo/bar:meow',
            username=None,
            password=None)
        get_mock.assert_has_calls([
            mock.call(mock.ANY, 'https://localhost/v2/', timeout=60),
            mock.call(mock.ANY, 'https://foo/bar',
                      params={'scope': 'repository:foo/bar:pull'},
                      auth=mock.ANY, timeout=60)])
        self.assertIsNotNone(self.client._cached_auth)
        self.assertEqual('bearer me0w',
                         self.client.session.headers['Authorization'])
        token_mock.assert_called_once_with('foo')

    @mock.patch.object(hashlib, 'new', autospec=True)
    def test_download_blob_from_manifest(self, mock_hash, get_mock):
        CONF.set_override('secure_cdn_registries', ['localhost'], group='oci')
        self.client.session.headers = {'Authorization': 'bearer zoo'}
        mock_file = mock.MagicMock(spec=io.BytesIO)
        mock_hash.return_value.hexdigest.side_effect = iter([
            ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f'
             '61caaff8a'),
            ('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e8'
             '86266e7ae')
        ])
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_1 = mock.Mock()
        get_1.status_code = 200
        manifest = {
            'layers': [{
                'digest': ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706'
                           '483bfa0f98a5e886266e7ae')}]
        }
        get_1.text = json.dumps(manifest)
        get_2 = mock.Mock()
        get_2.status_code = 301
        get_2.headers = {'Location': 'https://localhost/foo/sha'}
        get_3 = mock.Mock()
        get_3.status_code = 200
        get_3.iter_content.return_value = ['some', 'content']
        get_mock.side_effect = iter([get_1, get_2, get_3])

        res = self.client.download_blob_from_manifest(
            'oci://localhost/foo/bar@sha256:' + csum,
            mock_file)

        mock_file.write.assert_has_calls([
            mock.call('some'),
            mock.call('content')])
        self.assertEqual(
            ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98'
             'a5e886266e7ae'),
            res)
        get_mock.assert_has_calls([
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/manifests/sha256:44136fa355b'
                 '3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a'),
                headers={
                    'Accept': 'application/vnd.oci.image.manifest.v1+json'},
                timeout=60),
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/blobs/sha256:2c26b46b68ffc68f'
                 'f99b453c1d30413413422d706483bfa0f98a5e886266e7ae'),
                stream=True,
                timeout=60),
            mock.call(
                mock.ANY,
                'https://localhost/foo/sha',
                stream=True,
                timeout=60)
        ])

    @mock.patch.object(hashlib, 'new', autospec=True)
    def test_download_blob_from_manifest_code_check(self, mock_hash,
                                                    get_mock):
        mock_file = mock.MagicMock(spec=io.BytesIO)
        mock_hash.return_value.hexdigest.side_effect = iter([
            ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f'
             '61caaff8a'),
            ('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e8'
             '86266e7ae')
        ])
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_1 = mock.Mock()
        get_1.status_code = 200
        manifest = {
            'layers': [{
                'digest': ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706'
                           '483bfa0f98a5e886266e7ae')}]
        }
        get_1.text = json.dumps(manifest)
        get_2 = mock.Mock()
        get_2.status_code = 301
        get_2.headers = {'Location': 'https://localhost/foo/sha'}
        get_3 = mock.Mock()
        get_3.status_code = 204
        get_3.iter_content.return_value = ['some', 'content']
        get_mock.side_effect = iter([get_1, get_2, get_3])

        self.assertRaisesRegex(
            exception.ImageRefValidationFailed,
            'Got HTTP code 204',
            self.client.download_blob_from_manifest,
            'oci://localhost/foo/bar@sha256:' + csum,
            mock_file)

        mock_file.write.assert_not_called()
        get_mock.assert_has_calls([
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/manifests/sha256:44136fa355b'
                 '3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a'),
                headers={
                    'Accept': 'application/vnd.oci.image.manifest.v1+json'},
                timeout=60),
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/blobs/sha256:2c26b46b68ffc68f'
                 'f99b453c1d30413413422d706483bfa0f98a5e886266e7ae'),
                stream=True,
                timeout=60),
            mock.call(
                mock.ANY,
                'https://localhost/foo/sha',
                stream=True,
                timeout=60)
        ])

    @mock.patch.object(hashlib, 'new', autospec=True)
    def test_download_blob_from_manifest_code_401(self, mock_hash,
                                                  get_mock):
        self.client.session.headers = {'Authorization': 'bearer zoo'}
        mock_file = mock.MagicMock(spec=io.BytesIO)
        mock_hash.return_value.hexdigest.side_effect = iter([
            ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f'
             '61caaff8a'),
            ('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e8'
             '86266e7ae')
        ])
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_1 = mock.Mock()
        get_1.status_code = 200
        manifest = {
            'layers': [{
                'digest': ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706'
                           '483bfa0f98a5e886266e7ae')}]
        }
        get_1.text = json.dumps(manifest)
        get_2 = mock.Mock()
        get_2.status_code = 401
        get_2_exc = requests.exceptions.HTTPError(
            response=get_2)
        # Authentication is automatically retried, hence
        # needing to return exceptions twice.
        get_mock.side_effect = iter([get_1, get_2_exc, get_2_exc])

        self.assertRaises(
            exception.ImageServiceAuthenticationRequired,
            self.client.download_blob_from_manifest,
            'oci://localhost/foo/bar@sha256:' + csum,
            mock_file)

        mock_file.write.assert_not_called()
        get_mock.assert_has_calls([
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/manifests/sha256:44136fa355b'
                 '3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a'),
                headers={
                    'Accept': 'application/vnd.oci.image.manifest.v1+json'},
                timeout=60),
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/blobs/sha256:2c26b46b68ffc68f'
                 'f99b453c1d30413413422d706483bfa0f98a5e886266e7ae'),
                stream=True,
                timeout=60),
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/blobs/sha256:2c26b46b68ffc68f'
                 'f99b453c1d30413413422d706483bfa0f98a5e886266e7ae'),
                stream=True,
                timeout=60),
        ])

    @mock.patch.object(hashlib, 'new', autospec=True)
    def test_download_blob_from_manifest_code_404(self, mock_hash,
                                                  get_mock):
        mock_file = mock.MagicMock(spec=io.BytesIO)
        mock_hash.return_value.hexdigest.side_effect = iter([
            ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f'
             '61caaff8a'),
            ('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e8'
             '86266e7ae')
        ])
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_1 = mock.Mock()
        get_1.status_code = 200
        manifest = {
            'layers': [{
                'digest': ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706'
                           '483bfa0f98a5e886266e7ae')}]
        }
        get_1.text = json.dumps(manifest)
        get_2 = mock.Mock()
        get_2.status_code = 404
        get_2_exc = requests.exceptions.HTTPError(
            response=get_2)
        get_mock.side_effect = iter([get_1, get_2_exc])

        self.assertRaises(
            exception.ImageNotFound,
            self.client.download_blob_from_manifest,
            'oci://localhost/foo/bar@sha256:' + csum,
            mock_file)

        mock_file.write.assert_not_called()
        get_mock.assert_has_calls([
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/manifests/sha256:44136fa355b'
                 '3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a'),
                headers={
                    'Accept': 'application/vnd.oci.image.manifest.v1+json'},
                timeout=60),
            mock.call(
                mock.ANY,
                ('https://localhost/v2/foo/bar/blobs/sha256:2c26b46b68ffc68f'
                 'f99b453c1d30413413422d706483bfa0f98a5e886266e7ae'),
                stream=True,
                timeout=60),
        ])

    @mock.patch.object(hashlib, 'new', autospec=True)
    def test_download_blob_from_manifest_code_403(self, mock_hash,
                                                  get_mock):
        mock_file = mock.MagicMock(spec=io.BytesIO)
        mock_hash.return_value.hexdigest.side_effect = iter([
            ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f'
             '61caaff8a'),
            ('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e8'
             '86266e7ae')
        ])
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_1 = mock.Mock()
        get_1.status_code = 200
        manifest = {
            'layers': [{
                'digest': ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706'
                           '483bfa0f98a5e886266e7ae')}]
        }
        get_1.text = json.dumps(manifest)
        get_2 = mock.Mock()
        get_2.status_code = 403
        get_2_exc = requests.exceptions.HTTPError(
            response=get_2)
        get_mock.side_effect = iter([get_1, get_2_exc])

        self.assertRaises(
            exception.ImageNotFound,
            self.client.download_blob_from_manifest,
            'oci://localhost/foo/bar@sha256:' + csum,
            mock_file)
        mock_file.write.assert_not_called()
        self.assertEqual(2, get_mock.call_count)

    @mock.patch.object(hashlib, 'new', autospec=True)
    def test_download_blob_from_manifest_code_500(self, mock_hash,
                                                  get_mock):
        mock_file = mock.MagicMock(spec=io.BytesIO)
        mock_hash.return_value.hexdigest.side_effect = iter([
            ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f'
             '61caaff8a'),
            ('2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e8'
             '86266e7ae')
        ])
        csum = ('44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c0'
                '60f61caaff8a')
        get_1 = mock.Mock()
        get_1.status_code = 200
        manifest = {
            'layers': [{
                'digest': ('sha256:2c26b46b68ffc68ff99b453c1d30413413422d706'
                           '483bfa0f98a5e886266e7ae')}]
        }
        get_1.text = json.dumps(manifest)
        get_2 = mock.Mock()
        get_2.status_code = 500
        get_2_exc = requests.exceptions.HTTPError(
            response=get_2)
        get_mock.side_effect = iter([get_1, get_2_exc])

        self.assertRaises(
            exception.TemporaryFailure,
            self.client.download_blob_from_manifest,
            'oci://localhost/foo/bar@sha256:' + csum,
            mock_file)
        mock_file.write.assert_not_called()
        self.assertEqual(2, get_mock.call_count)


class TestRegistrySessionHelper(base.TestCase):

    def test_get_token_from_config_default(self):
        self.assertIsNone(
            oci_registry.RegistrySessionHelper.get_token_from_config(
                'foo.bar'))

    def test_get_token_from_config(self):
        CONF.set_override('authentication_config', '/tmp/foo',
                          group='oci')
        read_data = """{"auths": {"foo.fqdn": {"auth": "secret"}}}"""
        with mock.patch('builtins.open', mock.mock_open(
                read_data=read_data)):
            res = oci_registry.RegistrySessionHelper.get_token_from_config(
                'foo.fqdn')
        self.assertEqual('secret', res)

    def test_get_token_from_config_no_match(self):
        CONF.set_override('authentication_config', '/tmp/foo',
                          group='oci')
        read_data = """{"auths": {"bar.fqdn": {}}}"""
        with mock.patch('builtins.open', mock.mock_open(
                read_data=read_data)):
            res = oci_registry.RegistrySessionHelper.get_token_from_config(
                'foo.fqdn')
        self.assertIsNone(res)

    def test_get_token_from_config_bad_file(self):
        CONF.set_override('authentication_config', '/tmp/foo',
                          group='oci')
        read_data = """{"auths":..."""
        with mock.patch('builtins.open', mock.mock_open(
                read_data=read_data)):
            res = oci_registry.RegistrySessionHelper.get_token_from_config(
                'foo.fqdn')
        self.assertIsNone(res)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""remove_extra_fk_constraint in allocations

Revision ID: d163df1bab88
Revises: 163040c5513f
Create Date: 2023-06-23 11:20:10.904262

"""


from alembic import op
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade


# revision identifiers, used by Alembic.
revision = 'd163df1bab88'
down_revision = '163040c5513f'


def upgrade():
    engine = enginefacade.reader.get_engine()

    if engine.dialect.name == 'mysql':
        # Remove the outer level conductor_affinity constraint which pointed
        # to the conductor.id field, when the allocation's table also already
        # points to unique node constraints where a node also has a conductor
        # affinity field on the conductor.id. Removing because this is
        # expected to be come a hard error for SQLAlchemy at some point.
        try:
            op.drop_constraint('allocations_ibfk_1', 'allocations',
                               type_="foreignkey")
            op.drop_constraint('allocations_ibfk_2', 'allocations',
                               type_="foreignkey")
        except db_exc.DBNonExistentConstraint:
            # This is to ignore this issue on newer deployments where
            # key mappings may not already exist.
            pass



# Copyright 2016 Intel Corporation
# Copyright 2014 Rackspace, Inc.
# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic.common.i18n import _


opts = [
    cfg.StrOpt('backend',
               default='noop',
               choices=[
                   ('noop', 'Do nothing in relation to metrics.'),
                   ('statsd', 'Transmits metrics data to a statsd backend.'),
                   ('collector', 'Collects metrics data and saves it in '
                                 'memory for use by the running application.'),
               ],
               help='Backend to use for the metrics system.'),
    cfg.BoolOpt('prepend_host',
                default=False,
                help='Prepend the hostname to all metric names. '
                     'The format of metric names is '
                     '[global_prefix.][host_name.]prefix.metric_name.'),
    cfg.BoolOpt('prepend_host_reverse',
                default=True,
                help='Split the prepended host value by "." and reverse it '
                     '(to better match the reverse hierarchical form of '
                     'domain names).'),
    cfg.StrOpt('global_prefix',
               help='Prefix all metric names with this value. '
                    'By default, there is no global prefix. '
                    'The format of metric names is '
                    '[global_prefix.][host_name.]prefix.metric_name.'),
    # IPA config options: used by IPA to configure how it reports metric data
    cfg.StrOpt('agent_backend',
               default='noop',
               help=_('Backend for the agent ramdisk to use for metrics. '
                      'Default possible backends are "noop" and "statsd".')),
    cfg.BoolOpt('agent_prepend_host',
                default=False,
                help=_('Prepend the hostname to all metric names sent by the '
                       'agent ramdisk. The format of metric names is '
                       '[global_prefix.][uuid.][host_name.]prefix.'
                       'metric_name.')),
    cfg.BoolOpt('agent_prepend_uuid',
                default=False,
                help=_('Prepend the node\'s Ironic uuid to all metric names '
                       'sent by the agent ramdisk. The format of metric names '
                       'is [global_prefix.][uuid.][host_name.]prefix.'
                       'metric_name.')),
    cfg.BoolOpt('agent_prepend_host_reverse',
                default=True,
                help=_('Split the prepended host value by "." and reverse it '
                       'for metrics sent by the agent ramdisk (to better '
                       'match the reverse hierarchical form of domain '
                       'names).')),
    cfg.StrOpt('agent_global_prefix',
               help=_('Prefix all metric names sent by the agent ramdisk '
                      'with this value. The format of metric names is '
                      '[global_prefix.][uuid.][host_name.]prefix.'
                      'metric_name.'))
]


statsd_opts = [
    cfg.StrOpt('statsd_host',
               default='localhost',
               help='Host for use with the statsd backend.'),
    cfg.PortOpt('statsd_port',
                default=8125,
                help='Port to use with the statsd backend.'),
    cfg.StrOpt('agent_statsd_host',
               default='localhost',
               help=_('Host for the agent ramdisk to use with the statsd '
                      'backend. This must be accessible from networks the '
                      'agent is booted on.')),
    cfg.PortOpt('agent_statsd_port',
                default=8125,
                help=_('Port for the agent ramdisk to use with the statsd '
                       'backend.')),
]


def register_opts(conf):
    conf.register_opts(opts, group='metrics')
    conf.register_opts(statsd_opts, group='metrics_statsd')


def list_opts():
    return [opts, statsd_opts]



# Copyright 2023 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pecan import rest

from ironic import api
from ironic.api.controllers.v1 import utils as api_utils
from ironic.api import method
from ironic.common import args
from ironic.common import metrics_utils
from ironic import objects

METRICS = metrics_utils.get_metrics_logger(__name__)

_DEFAULT_RETURN_FIELDS = ('component', 'initial_version', 'current_version',
                          'last_version_flashed')


# NOTE(iurygregory): Keeping same parameters just in case we decide
# to support /v1/nodes/<node_uuid>/firmware/<component>
def convert_with_links(rpc_firmware, node_uuid, detail=None, fields=None):
    """Build a dict containing a firmware component."""

    fw_component = api_utils.object_to_dict(
        rpc_firmware,
        include_uuid=False,
        fields=fields,
    )
    return fw_component


def collection_from_list(node_ident, firmware_components, detail=None,
                         fields=None):
    firmware_list = []
    for fw_cmp in firmware_components:
        firmware_list.append(convert_with_links(fw_cmp, node_ident,
                             detail, fields))
    return {'firmware': firmware_list}


class NodeFirmwareController(rest.RestController):
    """REST controller for Firmware."""

    def __init__(self, node_ident=None):
        super(NodeFirmwareController, self).__init__()
        self.node_ident = node_ident

    @METRICS.timer('NodeFirmwareController.get_all')
    @method.expose()
    @args.validate(fields=args.string_list, detail=args.boolean)
    def get_all(self, detail=None, fields=None):
        """List node firmware components."""
        node = api_utils.check_node_policy_and_retrieve(
            'baremetal:node:firmware:get', self.node_ident)

        allow_query = api_utils.allow_firmware_interface
        fields = api_utils.get_request_return_fields(fields, detail,
                                                     _DEFAULT_RETURN_FIELDS,
                                                     allow_query, allow_query)
        components = objects.FirmwareComponentList.get_by_node_id(
            api.request.context, node.id)
        return collection_from_list(self.node_ident, components,
                                    detail, fields)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from ironic.common import exception
from ironic.common import keystone
from ironic.tests.base import TestCase


class KeystoneTestCase(TestCase):

    def setUp(self):
        super(KeystoneTestCase, self).setUp()
        self.test_group = 'test_group'
        self.cfg_fixture.conf.register_group(cfg.OptGroup(self.test_group))
        keystone.register_auth_opts(self.cfg_fixture.conf, self.test_group,
                                    service_type='vikings')
        self.config(auth_type='password',
                    group=self.test_group)
        # NOTE(pas-ha) this is due to auth_plugin options
        # being dynamically registered on first load,
        # but we need to set the config before
        plugin = ks_loading.get_plugin_loader('password')
        opts = ks_loading.get_auth_plugin_conf_options(plugin)
        self.cfg_fixture.register_opts(opts, group=self.test_group)
        self.config(auth_url='http://127.0.0.1:9898',
                    username='fake_user',
                    password='fake_pass',
                    project_name='fake_tenant',
                    group=self.test_group)

    def test_get_session(self):
        self.config(timeout=10, group=self.test_group)
        session = keystone.get_session(self.test_group, timeout=20)
        self.assertEqual(20, session.timeout)

    def test_get_auth(self):
        auth = keystone.get_auth(self.test_group)
        self.assertEqual('http://127.0.0.1:9898', auth.auth_url)

    def test_get_auth_fail(self):
        # NOTE(pas-ha) 'password' auth_plugin is used,
        # so when we set the required auth_url to None,
        # MissingOption is raised
        self.config(auth_url=None, group=self.test_group)
        self.assertRaises(exception.ConfigInvalid,
                          keystone.get_auth,
                          self.test_group)

    def test_get_adapter_from_config(self):
        self.config(valid_interfaces=['internal', 'public'],
                    group=self.test_group)
        session = keystone.get_session(self.test_group)
        adapter = keystone.get_adapter(self.test_group, session=session,
                                       interface='admin')
        self.assertEqual('admin', adapter.interface)
        self.assertEqual(session, adapter.session)

    @mock.patch('keystoneauth1.service_token.ServiceTokenAuthWrapper',
                autospec=True)
    @mock.patch('keystoneauth1.token_endpoint.Token', autospec=True)
    def test_get_service_auth(self, token_mock, service_auth_mock):
        ctxt = mock.Mock(spec=['auth_token'], auth_token='spam')
        mock_auth = mock.Mock()
        self.assertEqual(service_auth_mock.return_value,
                         keystone.get_service_auth(ctxt, 'ham', mock_auth))
        token_mock.assert_called_once_with('ham', 'spam')
        service_auth_mock.assert_called_once_with(
            user_auth=token_mock.return_value, service_auth=mock_auth)


class AuthConfTestCase(TestCase):

    def setUp(self):
        super(AuthConfTestCase, self).setUp()
        self.test_group = 'test_group'
        self.cfg_fixture.conf.register_group(cfg.OptGroup(self.test_group))
        keystone.register_auth_opts(self.cfg_fixture.conf, self.test_group)
        self.config(auth_type='password',
                    group=self.test_group)
        # NOTE(pas-ha) this is due to auth_plugin options
        # being dynamically registered on first load,
        # but we need to set the config before
        plugin = ks_loading.get_plugin_loader('password')
        opts = ks_loading.get_auth_plugin_conf_options(plugin)
        self.cfg_fixture.register_opts(opts, group=self.test_group)
        self.config(auth_url='http://127.0.0.1:9898',
                    username='fake_user',
                    password='fake_pass',
                    project_name='fake_tenant',
                    group=self.test_group)

    def test_add_auth_opts(self):
        opts = keystone.add_auth_opts([])
        # check that there is no duplicates
        names = {o.dest for o in opts}
        self.assertEqual(len(names), len(opts))
        # NOTE(pas-ha) checking for most standard auth and session ones only
        expected = {'timeout', 'insecure', 'cafile', 'certfile', 'keyfile',
                    'auth_type', 'auth_url', 'username', 'password',
                    'tenant_name', 'project_name', 'trust_id',
                    'domain_id', 'user_domain_id', 'project_domain_id'}
        self.assertTrue(expected.issubset(names))

    def test_os_service_types_alias(self):
        keystone.register_auth_opts(self.cfg_fixture.conf, 'barbican')
        self.assertEqual(self.cfg_fixture.conf.barbican.service_type,
                         'key-manager')
